import socket
from pathlib import Path
import typer
import os

APP_NAME = "remla"
hostname = socket.gethostname()
packagesToCheck = ["nginx", "python3-pip", "i2c-tools", "pigpio"]




###### List of important paths (for now)
mediaMTX_tar_file = "https://github.com/bluenviron/mediamtx/releases/download/v1.9.3/mediamtx_v1.9.3_linux_arm64v8.tar.gz"
mediamtxVersion = "1.9.3"
mediamtxSettingsLocation = Path("/usr/local/etc")
mediamtxBinaryLocation = Path("/usr/local/bin")
baseDir = Path(__file__).parent
settingsDirectory = Path(typer.get_app_dir(APP_NAME))
logsDirectory = settingsDirectory / "logs"
homeDirectory = Path.home()
remoteLabsDirectory = homeDirectory / 'remla'
setupDirectory = baseDir / "setup"
websiteDirectory = settingsDirectory / 'website'
nginxTemplatePath = setupDirectory / "remla.conf"
nginxConfPath = Path("/etc/nginx/sites-available/remla.conf")
nginxConfLinkPath = Path("/etc/nginx/sites-enabled/remla.conf")
nginxAvailablePath = Path("/etc/nginx/sites-available")
nginxEnabledPath = Path("/etc/nginx/sites-enabled")
localhostConfLinkPath = nginxEnabledPath / "localhost.conf"
bootConfigPath = Path("/boot/firmware/config.txt")
nginxWebsitePath = Path("/var/www/remla")
pidFilePath = Path(f'/var/run/user/{str(os.environ.get("SUDO_UID")) if os.environ.get("SUDO_UID") is not None else str(os.getuid())}/remla.pid')
websiteStaticDirectory = websiteDirectory / "static"
websiteJSDirectory = websiteStaticDirectory / "js"
websiteCSSDirectory = websiteStaticDirectory / "css"
websiteImgsDirectory = websiteStaticDirectory / "imgs"



#!/home/remoteLabs/.cache/pypoetry/virtualenvs/remla-BlFEVOYb-py3.11/bin/python

import asyncio
import datetime
import os
import re
import shutil
import signal
import socket
import subprocess
from pathlib import Path
from typing import Optional

import typer
import websockets
from rich import print as rprint
from rich.markdown import Markdown
from rich.prompt import IntPrompt
from rich.text import Text
from typing_extensions import Annotated

from remla import i2ccmd, setupcmd
from remla.labcontrol.Controllers import *
from remla.labcontrol.Experiment import Experiment
from remla.settings import *
from remla.systemHelpers import *
from remla.typerHelpers import *
from remla.yaml import createDevicesFromYml, yaml

from .customvalidators import *

__version__ = "0.1.4"


def version_callback(value: bool):
    if value:
        print(f"Remla Version: {__version__}")
        raise typer.Exit()


app = typer.Typer()
app.add_typer(setupcmd.app, name="setup")
app.add_typer(i2ccmd.app, name="i2c")


@app.callback()
def version(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        is_eager=True,
        help="Print the version and exit",
    ),
):
    pass


@app.command()
def showconfig():
    app_dir = typer.get_app_dir(APP_NAME)
    typer.echo(app_dir)


@app.command(
    help="Run this to initilize your remla setup. It will make sure you have "
    "the correct dependencies installed, as well as the install mediamtx for "
    "you. It will then then take you through setting up your hardware. Just "
    "answer the questions and it will setup everything for you"
)
def init():
    ####### Verify that we are running as sudo ######
    if os.geteuid() != 0:
        alert("This script must be run as root.")
        typer.echo("Try running:")
        typer.echo("sudo remla init")
        raise typer.Abort()

    ####### Make config directory and others #########
    settingsDirectory.mkdir(parents=True, exist_ok=True)
    logsDirectory.mkdir(parents=True, exist_ok=True)
    websiteDirectory.mkdir(parents=True, exist_ok=True)
    websiteStaticDirectory.mkdir(parents=True, exist_ok=True)
    websiteJSDirectory.mkdir(parents=True, exist_ok=True)
    websiteCSSDirectory.mkdir(parents=True, exist_ok=True)
    websiteImgsDirectory.mkdir(parents=True, exist_ok=True)
    remoteLabsDirectory.mkdir(parents=True, exist_ok=True)

    ####### Enable I2C #######
    remlaPanel("Turning on I2C on your raspberry pi.")
    try:
        subprocess.run(
            ["sudo", "raspi-config", "nonint", "do_i2c", "0"], check=True
        )  # 0 mean true in bash I guess
        success("Turned on I2C")
    except subprocess.CalledProcessError as e:
        alert(f"Failed to turn on I2C with error {e}")
        raise typer.Abort()

    ####### Check to see if apt required packages are installed.  ###########
    remlaPanel("Verifying required apt packages are installed.")
    packagesToCheck = ["nginx", "python3-pip", "i2c-tools", "pigpio", "python3-pigpio"]
    packagesNeeded = []
    for package in packagesToCheck:
        if not is_package_installed(package):
            packagesNeeded.append(package)
    if len(packagesNeeded) != 0:
        alert("You have missing required packages!")
        typer.echo("Please run the following command:")
        typer.echo(f"sudo apt install {' '.join(packagesNeeded)}")
        raise typer.Abort()
    success("All required packages are installed!")

    ###### Turn on PIGPIOD to run at start
    typer.echo("Turning on pigpiod to start at boot.")
    echoResult(
        enable_service("pigpiod"),
        "pigiod will start at boot",
    )

    ####### Installing mediamtx for the user by downloading it from github, unpacking it, and moving files.
    _mediamtx()

    ####### Seting up NGINX for the user now ############
    _nginx()

    ####### Create an initial settings file #############
    _createSettingsFile()
    ####### Create a remla.service daemon   #############
    createServiceFile(echo=True)
    createRemlaPolicy()

    interactivesetup()
    typer.echo("Wrapping up install...")
    subprocess.run(["sudo", "systemctl", "daemon-reload"])
    subprocess.run(["sudo", "systemctl", "restart", "remla.service"])
    enable_service("remla")
    # alert("Running test websocket server. Press Ctrl-C when you are done testing.")
    # run(wstest=True)


def _mediamtx():
    remlaPanel("Installing MediaMTX")
    typer.echo("  Checking for prior installation")
    mediamtxInstalled = False
    if os.path.exists("/usr/local/bin/mediamtx"):
        typer.echo("  Already found Mediamtx Installation")
        mediamtxInstalled = True
    else:
        echoResult(
            download_and_extract_tar(mediaMTX_tar_file, settingsDirectory, "mediamtx"),
            "Downloaded and extracted MediaMTX",
            "Something went wrong in the downloading and extracting process. Check internet and try again.",
        )
    typer.echo("  Creating MediaMTX Systemlinks to fix LibCameraBug")
    typer.echo("  Creating MediaMTX settings file")
    # Change log file location in mediamtx.yml settings file.
    # Then save the new mediamtx.yml file to /usr/local/etc where mediamtx says to locate
    # the file.
    mediamtxSettings = yaml.load(setupDirectory / "mediamtx.yml")
    mediamtxSettings["logFile"] = str(logsDirectory / "mediamtx.log")
    encryptionValue = mediamtxSettings["encryption"]
    rtmpEncryptionValue = mediamtxSettings["rtmpEncryption"]
    mediamtxSettings["encryption"] = "<replace1>"
    mediamtxSettings["rtmpEncryption"] = "<replace2>"
    mediamtxSettingsLocation = Path("/usr/local/etc")
    (mediamtxSettingsLocation / "mediamtx.yml").unlink(missing_ok=True)

    mediamtxSettingsLocation.mkdir(parents=True, exist_ok=True)
    yaml.dump(mediamtxSettings, mediamtxSettingsLocation / "mediamtx.yml")
    with open(mediamtxSettingsLocation / "mediamtx.yml", "r") as file:
        content = file.read()
    content = content.replace("<replace1>", f'"{encryptionValue}"')
    content = content.replace("<replace2>", f'"{rtmpEncryptionValue}"')
    with open(mediamtxSettingsLocation / "mediamtx.yml", "w") as file:
        file.write(content)

    # Now move mediamtx binary to /usr/local/bin where mediamtx says to move it
    if not mediamtxInstalled:
        moveAndOverwrite(
            settingsDirectory / "mediamtx/mediamtx", mediamtxBinaryLocation
        )
    # Move service file to systemd so that we can run it on boot.
    shutil.copy(setupDirectory / "mediamtx.service", "/etc/systemd/system")

    # Finally setup systemd to run this service on start.
    subprocess.run(["sudo", "systemctl", "daemon-reload"])
    subprocess.run(["sudo", "systemctl", "enable", "mediamtx"])
    subprocess.run(["sudo", "systemctl", "start", "mediamtx"])
    subprocess.run(["sudo", "systemctl", "restart", "mediamtx"])
    success("Successfully set up mediamtx")


def _nginx():
    logsDirectory.mkdir(parents=True, exist_ok=True)
    typer.echo("Setting up NGINX")
    # Make directory for the running website.
    # nginxWebsitePath.mkdir(parents=True, exist_ok=True)

    shutil.copy(setupDirectory / "reader.js", websiteJSDirectory)
    shutil.copy(setupDirectory / "mediaMTXGetFeed.js", websiteJSDirectory)

    updateRemlaNginxConf(8080, hostname, 8675)

    updatedHtml = updateFinalInfo(setupDirectory / "index.html")
    # Write the processed HTML to a new file or use as needed
    with open(websiteDirectory / "index.html", "w") as file:
        file.write(updatedHtml)

    shutil.copytree(websiteDirectory, nginxWebsitePath, dirs_exist_ok=True)

    typer.echo("Making NGINX run at boot")
    echoResult(
        enable_service("nginx"),
        "NGINX can now run at boot.",
        "Failed to allow NGINX to start at boot.",
    )
    subprocess.run(["sudo", "systemctl", "reload", "nginx"])
    success("NGINX setup complete.")
    # Change Permission so NGINX can access files
    # homeDirectory.chmod(0o755)


@app.command()
def interactivesetup():
    user = homeDirectory.owner()
    message = "Note that remla currently only works with Raspberry Pi 4! If you are using a newer model, you will need do this manually."
    remlaPanel(message)
    (cont_int,) = (
        typer.confirm("Do you want to continue with interactive install?", default="y"),
    )
    if not cont_int:
        return
    allowedSensors = ["ov5647", "imx219", "imx477", "imx708", "imx519", "other"]
    sensorQuestionString = "Select which type of sensor you will be using [1-5]:\n"
    for i, sensor in enumerate(allowedSensors):
        sensorQuestionString += f"  {i+1}. {sensor} \n"
    sensorIdx = (
        IntPrompt.ask(sensorQuestionString, choices=["1", "2", "3", "4", "5", "6"]) - 1
    )
    customSensor = False
    if sensorIdx == 5:
        customSensor = True
        sensor = typer.prompt(
            "Please provide the name of your sensor as required by raspberry pi or arducam (make sure you know what you are doing)",
            confirmation_prompt=True,
        )
    else:
        sensor = allowedSensors[sensorIdx]

    while True:
        numCameras = IntPrompt.ask(
            "How many cameras will you be using?", choices=["1", "2", "3", "4"]
        )
        cameraChoices = {
            2: {"1": "A", "2": "B"},
            3: {"1": "A", "2": "B", "3": "C", "4": "D"},
        }
        cameraPorts = []
        newline_space = "\n "
        multiplexerQuestion = "Which type of multiplexer are you using:\n  1. None\n  2. Arducam 2 Camera Multiplexer\n  3. Arducam 4 Camera Mutiplexer\n"
        multiplexer = IntPrompt.ask(multiplexerQuestion, choices=["1", "2", "3"])
        if multiplexer == 3 or numCameras <= multiplexer:
            if multiplexer in [2, 3]:
                for i in range(numCameras):
                    prompt = f"Which port is camera {i+1} connected to:\n {newline_space.join(f'{num}. {port}' for num, port in cameraChoices[multiplexer].items())}\n"
                    port = IntPrompt.ask(
                        prompt, choices=list(cameraChoices[multiplexer].keys())
                    )
                    cameraChoices[multiplexer].pop(str(port))
                    cameraPorts.append(int(port) - 1)
            break

        else:
            warning(
                "There is a discrepancy between your multiplexer choice and the number of cameras you have.\n"
                " You can't have more cameras than slots for cameras.\n Starting again."
            )

    remlaPanel("Now updating /boot/firmware/config.txt")
    arducamMultiplexers = {2: "camera-mux-2port", 3: "camera-mux-4port"}
    dtOverlayString = "dtoverlay="

    if multiplexer == 1:
        dtOverlayString += sensor
    else:
        cams = ["cam" + str(i) + "-" + sensor for i in cameraPorts]
        arducamString = ",".join(cams)
        dtOverlayString += f"{arducamMultiplexers[multiplexer]},{arducamString}"

    if customSensor:
        warning("Issue with custom sensor!")
        rprint(
            f"Because you provided the custom sensor, [green]{sensor}[/green], it is not guaranteed this installer"
            f" can successfully make the changes to /boot/firmware/config.txt"
        )
        rprint(
            f"Therefore you will need to manually add the line \n[i green]{dtOverlayString}[/i green] \nto your "
            f"config.txt. Just make sure that you don't have two camera related dtoverlays."
        )
        rprint(
            "To make the change, copy the dtoverlay string above, run the command \n[i green]sudo nano "
            "/boot/firmware/config.txt[/i green] \nand paste over your previous dtoverlay or right "
            "below the camera_auto_detect line."
        )

    else:
        with open(bootConfigPath, "r") as file:
            config = file.readlines()

        # Additional setup for camera_auto_detect replacement
        cameraAutoDetectSearch = "camera_auto_detect=1"
        cameraAutoDetectReplace = "camera_auto_detect=0"

        # Prepare the regex pattern
        # Combine allowed sensors and arducam multiplexer values into one list for the regex pattern
        combinedOptions = allowedSensors + list(arducamMultiplexers.values())
        pattern = re.compile(
            r"dtoverlay=("
            + "|".join(re.escape(option) for option in combinedOptions)
            + ")"
        )

        # Search and replace the line, or prepare to append
        dtOverlayFound = False
        for i, line in enumerate(config):
            # Replace camera_auto_detect line if found
            if cameraAutoDetectSearch in line:
                config[i] = cameraAutoDetectReplace + "\n"
                rprint(
                    f"Switching [red]{cameraAutoDetectSearch}[/red] --> [green]{cameraAutoDetectReplace}[/green]"
                )
            if pattern.search(line):
                rprint(
                    f"Switching [red]{config[i]}[/red] --> [green]{dtOverlayString}[/green]"
                )
                config[i] = dtOverlayString + "\n"  # Replace the line
                dtOverlayFound = True
                break

        # If the pattern wasn't found, append the dtOverlayString
        if not dtOverlayFound:
            config.append(dtOverlayString + "\n")
            rprint(
                f"Did not find any camera related dtoverlays in /boot/firmware/config.txt. Appending {dtOverlayString} to end of file."
            )

        # Write the modified content back to the config file
        with open(bootConfigPath, "w") as file:
            file.writelines(config)

        localip = _localip()

        finalInfo = updateFinalInfo(setupDirectory / "finalInfoTemplate.md")
        with open(settingsDirectory / "finalInfo.md", "w") as file:
            file.write(finalInfo)

        subprocess.run(
            ["sudo", "chown", "-R", f"{user}:{user}", f"{remoteLabsDirectory}"]
        )
        subprocess.run(
            ["sudo", "chown", "-R", f"{user}:{user}", f"{settingsDirectory}"]
        )

        message = Text(
            f"You have finished installing remla, the remoteLabs control center.\n"
            f"The next is for you to go one of:\n"
            f"http://{hostname}.local:8080\n"
            f"http://{localip}:8080\n"
            f"Follow the instructions there."
            f"If that doesn't work then run `remla finalinfo` to see it in the command line.",
            justify="center",
        )

        panelDisplay(
            message, title="🎉🎉🎉 Congratulations! 🎉🎉🎉", border_style="green"
        )


def _createSettingsFile():
    """
    Creates the settings file living in the settings folder.
    """

    settings = {
        "paths": {
            "settingsDirectory": settingsDirectory,
            "remoteLabsDirectory": remoteLabsDirectory,
            "websiteDirectory": websiteDirectory,
            "logsDirectory": logsDirectory,
            "nginxTemplatePath": nginxTemplatePath,
            "mediamtxBinaryLocation": mediamtxBinaryLocation,
            "mediamtxSettingsLocation": mediamtxSettingsLocation,
        },
        "currentLab": None,
    }

    with open(settingsDirectory / "settings.yml", "w") as file:
        yaml.dump(settings, file)


@app.command(help="Display finall installation info in the terminal")
def finalinfo():
    with open(settingsDirectory / "finalInfo.md", "r") as file:
        markdown = file.read()
        md = Markdown(markdown)
    rprint(md)


def _ip():
    try:
        # Create a dummy socket to connect to an external site
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            # Use Google's Public DNS server to find the best local IP
            s.connect(("8.8.8.8", 80))
            local_ip = s.getsockname()[0]
        rprint(local_ip)
    except Exception as e:
        print(f"Error obtaining local IP address: {e}")
        return None


@app.command()
def localip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(("10.254.254.254", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    typer.echo(IP)


def _localip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(("10.254.254.254", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


def updateFinalInfo(template: Path) -> str:
    """
    Taks in a file reads it as text. Makes the substitutions of placeholders and give the contents of the file back as a string.
    :param template: Path to file
    :return: string with updated text from file.
    """
    # Placeholder values
    placeholders = {
        "{{ remoteLabsDirectory }}": str(remoteLabsDirectory.name),
        "{{ settingsDirectory }}": str(settingsDirectory),
        "{{ packagesToCheck }}": ", ".join(packagesToCheck[1:]),
        "{{ mediamtxVersion }}": mediamtxVersion,
        "{{ mediamtxBinaryLocation }}": str(mediamtxBinaryLocation),
        "{{ mediamtxSettingsLocation }}": str(mediamtxSettingsLocation),
        "{{ nginxWebsitePath }}": str(nginxWebsitePath),
    }

    # Read the HTML template
    with open(template, "r") as file:
        content = file.read()

    # Replace each placeholder with its corresponding value
    for placeholder, replacement in placeholders.items():
        content = content.replace(placeholder, replacement)

    return content


@app.command("run")
@app.command("start")
def run(
    admin: Optional[bool] = typer.Option(False, "--admin", "-a", help="Run as admin."),
    foreground: Optional[bool] = typer.Option(
        False, "--foreground", "-f", help="Run in the foreground"
    ),
    wstest: Optional[bool] = typer.Option(
        False, "--wstest", "-w", help="Runs echo test server"
    ),
):
    now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print("#" * 80)
    print(f"########{now.center(64)}########")
    print("#" * 80)
    print()
    if status():
        warning(
            "Remla is already running. If you want to restart run `remla restart` or stop before running with new options."
        )
        raise typer.Abort()
    signal.signal(signal.SIGTERM, lambda signum, frame: cleanupPID())
    signal.signal(signal.SIGINT, lambda signum, frame: cleanupPID())

    if wstest:
        print("Starting Echo Server")

        async def echo(websocket, path):
            async for message in websocket:
                await websocket.send(f"Message received cap'n: {message}")

        start_server = websockets.serve(echo, "localhost", 8675)

        asyncio.get_event_loop().run_until_complete(start_server)
        asyncio.get_event_loop().run_forever()
    elif not foreground:
        try:
            subprocess.run(["systemctl", "start", "remla.service"], check=True)
            success("Running remla in background!")
        except subprocess.CalledProcessError as e:
            alert(f"Failed to start remla due to {e}")
            raise typer.Abort()
    else:
        print("Starting remla websocket server")
        typer.echo("Echo above statment")
        remlaSettingsPath = settingsDirectory / "settings.yml"
        # Check if the settings file exists
        if not remlaSettingsPath.exists():
            alert(f"Settings file not found at {remlaSettingsPath}.")
            raise typer.Abort()

        remlaSettings = yaml.load(remlaSettingsPath)
        currentLabSettingsPath = remoteLabsDirectory / remlaSettings["currentLab"]

        if not currentLabSettingsPath or not currentLabSettingsPath.exists():
            alert(
                f"Lab settings file does not exist or no current lab configured at {currentLabSettingsPath}. Please check your settings.yml."
            )
            raise typer.Abort()

        labSettings = yaml.load(currentLabSettingsPath)
        if "devices" not in labSettings:
            alert(
                f"Device list not found in the lab settings file located at {currentLabSettingsPath}. Please update the file to include your list of devices."
            )
            raise typer.Abort()

        # Initialize devices from the lab settings
        devices = createDevicesFromYml(labSettings["devices"])
        print("Using devices:", labSettings["devices"])
        # Create and setup the experiment
        if admin:
            experiment = Experiment("RemoteLabs", admin=True)
        else:
            experiment = Experiment("RemoteLabs")

        for device in devices.values():
            experiment.addDevice(device)

        #### Now set up the locks.
        locksConfig = labSettings.get("locks", {})

        for lockGroup, deviceNames in locksConfig.items():
            try:
                # Convert device names to device objects
                deviceObjects = [
                    devices[name] for name in deviceNames if name in devices
                ]

                # In case some devices listed in YAML are not initialized or missing
                if len(deviceObjects) != len(deviceNames):
                    missingDevices = set(deviceNames) - set(devices.keys())
                    alert(
                        f"Lock group '{lockGroup}' refers to undefined devices: {missingDevices}"
                    )
                    raise typer.Abort()

                # Apply the lock to the group of device objects
                experiment.addLock(deviceObjects)
            except KeyError as e:
                alert(f"Device name error in lock configuration: {str(e)}")
                raise typer.Abort()
        # Placeholder for further experiment execution logic
        success("Experiment setup complete.")
        experiment.startServer()


@app.command()
def stop():
    try:
        typer.echo(
            "Stopping remla. This could take some time for the system to reset to its starting parameters. Please be patient."
        )
        subprocess.run(["systemctl", "stop", "remla.service"], check=True)
        success("Stopped running remla")
    except subprocess.CalledProcessError:
        alert("Failed to stop remla")


@app.command()
def status():
    # pidFilePathFull = pidFilePath.replace("<uid>", str(getCallingUserID()))
    print(pidFilePath)
    if os.path.exists(pidFilePath):
        # Read exisitng pid file
        with open(pidFilePath, "r") as file:
            try:
                pid = int(file.read().strip())
                os.kill(pid, 0)
                typer.echo("Remla is already running")
                return True
            except ValueError:
                typer.echo("PID File is corrupt. Starting a new instance.")
            except ProcessLookupError:
                typer.echo("Remla instance not found. Staring new isntance")
            except PermissionError:
                typer.echo("Permission denied when checking PID. Assuming its running.")
                return True
    else:
        typer.echo("No PID file found. Starting new instance of remla")

    with open(pidFilePath, "w+") as file:
        file.write(str(os.getpid()))
    return False


@app.command()
def enable():
    try:
        subprocess.run(["systemctl", "enable", "remla.service"], check=True)
        success("Remla will now run on boot.")
    except subprocess.CalledProcessError:
        alert("Something went wrong.")


@app.command()
def disable():
    try:
        subprocess.run(["systemctl", "disable", "remla.service"], check=True)
        success("Remla will not run on boot.")
    except subprocess.CalledProcessError:
        alert("Something went wrong.")


def createRemlaPolicy():
    user = homeDirectory.owner()
    # Allows remla users to run
    policyKit = """[Allow Non-root Users to Manage Remla Service]
Identity=unix-group:remlausers
Action=org.freedesktop.systemd1.manage-units
ResultActive=yes
ResultInactive=yes
ResultAny=yes
"""
    groupName = "remlausers"
    with open(Path("/etc/polkit-1/localauthority/50-local.d/remla.pkla"), "w") as file:
        file.write(policyKit)
    try:
        subprocess.run(["sudo", "groupadd", groupName], check=True)
        success(f"Group '{groupName}' created successfully.")
    except subprocess.CalledProcessError:
        warning(f"Failed to create group '{groupName}'. It may already exist.")
    try:
        subprocess.run(["sudo", "usermod", "-a", "-G", groupName, user], check=True)
        success(f"User '{user}' added to group '{groupName}' successfully.")
    except subprocess.CalledProcessError:
        warning(f"Failed to add user '{user}' to group '{groupName}'.")


@app.command()
def git(giturl: Annotated[str, typer.Argument()]):
    if not urlValidator(giturl):
        raise typer.Abort()
    # Parse the URL to get the path component, specifically the last part
    repo_path = Path(giturl)
    base_name = repo_path.name

    # Remove the .git from the end if it exists
    if base_name.endswith(".git"):
        base_name = base_name[:-4]

    # Replace any characters not allowed in directory names, if necessary
    safe_name = re.sub(r"[^\w\-_\. ]", "_", base_name)
    cloneDirectory = remoteLabsDirectory / safe_name
    try:
        cloneDirectory.mkdir(exist_ok=False)
        subprocess.run(["git", "clone", giturl, cloneDirectory], check=True)
        success(f"Cloned directory to here to {cloneDirectory}")
    except FileExistsError:
        alert(
            f"That git repo already has a directory in {cloneDirectory}. Rename that folder before continuing"
        )
        raise typer.Abort()
    except subprocess.CalledProcessError as e:
        alert(f"There was an issue cloning the repo:\n{e}")
        raise typer.Abort()


@app.command()
def testws():
    async def echo(websocket, path):
        async for message in websocket:
            await websocket.send(message)

    start_server = websockets.serve(echo, "localhost", 8000)

    asyncio.get_event_loop().run_until_complete(start_server)
    asyncio.get_event_loop().run_forever()


if __name__ == "__main__":
    app()
# TODO: Create new command that builds a new lab.
# TODO: Create a setup command that shifts files around
# TODO: mediamtx just doesn't work right now for some reason.



import subprocess
import requests
import tarfile
import os
import functools
import typer
import sys

from remla.typerHelpers import *
from pathlib import Path
from rich.progress import track
from rich.prompt import IntPrompt
import shutil
from remla.settings import *
import re
from contextlib import contextmanager
from typing import Callable
from remla.customvalidators import *
from remla.yaml import yaml

def is_package_installed(package_name):
    try:
        # Attempt to show the package information
        # This is for Debian based OS's like Raspberry Pi Os
        subprocess.run(["dpkg", "-s", package_name], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        # The command failed, which likely means the package is not installed
        return False


def enable_service(service_name:str):
    try:
        subprocess.run(["sudo", "systemctl", "enable", service_name], check=True)
        success(f"Service {service_name} enabled successfully.")
        return True
    except subprocess.CalledProcessError as e:
        alert(f"Failed to enable {service_name}: {e}")
        return False

def download_and_extract_tar(url:str, savePath:Path, folderName:str):
    # Download the file
    response = requests.get(url, stream=True)
    # Check if the request was successful
    if response.status_code == 200:
        # Open a file to write the content of the download
        fileName = url.split("/")[-1]
        totalSizeInBytes = int(response.headers.get("content-length",0))
        saveLocation = savePath / fileName
        with open(saveLocation, "xb") as file:
            # Iterate over the response data in chunks (e.g., 4KB chunks)
            for chunk in track(response.iter_content(chunk_size=4096), description="Downloading...", total=totalSizeInBytes//4096):
                file.write(chunk)
        typer.echo("Extracting file...")
        # Extract the tar file
        extractPath = savePath / folderName
        with tarfile.open(saveLocation, "r:gz") as tar:
            tar.extractall(path=extractPath)
        # Optionally, remove the tar file after extraction
        os.remove(saveLocation)
        return True
    else:
        return False


def checkFileFullName(folderPath: Path, filePattern: str):
    """
    Check if a file matching a certain pattern exists in a specified folder.

    :param folderPath: Path to the folder where to search for the file.
    :param filePattern: The pattern to match the file names against.
    :return: Return first found matching file name, False otherwise.
    """
    folder = Path(folderPath)
    # Use glob to find matching files. This returns a generator.
    matchingFiles = folder.glob(filePattern)

    # Attempt to get the first matching file. If none exist, None is returned.
    try:
        firstMatch = next(matchingFiles)
        # If we get here, a matching file exists
        return firstMatch
    except StopIteration:
        # If no matching file exists, a StopIteration exception is caught
        return False

def moveAndOverwrite(source:Path, dest:Path):
    if dest.is_dir():
        (dest / source.name).unlink(missing_ok=True)
    else:
        dest.unlink(missing_ok=True)
    shutil.move(source, dest)

def getSettings():
    dir = Path(typer.get_app_dir(APP_NAME))
    # with open(dir, "r") as file:
    #     settingsString = file.read()

    return yaml.load(dir/"settings.yml")

def clearDirectory(directory: Path) -> None:
    if directory.exists() and directory.is_dir():
        for item in directory.iterdir():
            if item.is_dir():
                shutil.rmtree(item)  # Recursively remove directories
            else:
                item.unlink()  # Remove files
        # print(f"All files and directories removed from {directory}")
    else:
        print(f"The specified path {directory} is not a valid directory")

def searchForFilePattern(directory:Path, pattern:str, invalidMsg:tuple[str,Callable[[str],None]]|None=None, abort:bool=True) -> list:
    files = list(directory.rglob(pattern))
    numOfFiles = len(files)

    if numOfFiles == 0:
        if invalidMsg is not None:
            msgType = invalidMsg[1]
            msg = invalidMsg[0]
            msgType(msg)
        if abort:
            raise typer.Abort()
    return files



def promptForNumericFile(prompt:str, directory:Path, pattern:str, warnMsg:str|None=None, abort:bool=True) -> Path:
    files = searchForFilePattern(directory,pattern, (warnMsg, warning), abort)
    # msg = f"Multiple files with the same name found in {remoteLabsDirectory}. Lab names must be unqiue."
    # uniqueValidator(files, (msg, alert))
    numOfFiles = len(files)
    if not prompt.endswith("\n"):
        prompt += "\n"
    for i, file in enumerate(files):
        prompt += f"{i + 1}. {file.relative_to(directory)}\n"
    choice = IntPrompt.ask(prompt, choices=[str(i + 1) for i in range(numOfFiles)]) - 1
    return files[choice]

def updateRemlaNginxConf(port: int, domain:str, wsPort:int) -> None:
    nginxInitialConfPath = setupDirectory / "localhost.conf"
    # Read in the file
    with open(nginxInitialConfPath, "r") as file:
        nginxInitialConf = file.read()
    # Use re.sub() to replace all instances of {{ settingsDirectory }} with the settingsDirectory
    modifiedConf = re.sub(r'\{\{\s*settingsDirectory\s*\}\}', str(settingsDirectory), nginxInitialConf)
    modifiedConf = re.sub(r'\{\{\s*nginxWebsitePath\s*\}\}', str(nginxWebsitePath), modifiedConf)
    modifiedConf = re.sub(r'\{\{\s*port\s*\}\}', str(port), modifiedConf)
    modifiedConf = re.sub(r'\{\{\s*hostname\s*\}\}', domain, modifiedConf)
    modifiedConf = re.sub(r'\{\{\s*wsPort\s*\}\}', str(wsPort), modifiedConf)

    modifiedConfPath = settingsDirectory / "remla.conf"
    # with normalUserPrivileges():
    with open(modifiedConfPath, "w") as file:
        file.write(modifiedConf)
    # writeFileAsUser(modifiedConfPath, modifiedConf)
    nginxAvailableSymPath = nginxAvailablePath / "remla.conf"
    if not nginxAvailableSymPath.exists():
        nginxAvailableSymPath.symlink_to(modifiedConfPath)
    nginxEnableSymPath = nginxEnabledPath / "remla.conf"
    if not nginxEnableSymPath.exists():
        nginxEnableSymPath.symlink_to(nginxAvailableSymPath)

def runAsUser(func:callable, *args, **kwargs):
    currentUid = os.geteuid()
    os.setuid(1000)
    result = func(*args, **kwargs)
    os.seteuid(currentUid)
    return result

def writeFileAsUser(file:Path, contents:str):
    currentUid = os.geteuid()
    os.setuid(1000)
    with open(file, "w") as f:
        f.write(contents)
    os.seteuid(currentUid)

@contextmanager
def normalUserPrivileges():
    original_euid = os.geteuid()
    original_egid = os.getegid()
    normal_uid = 1000
    normal_gid = 1000

    try:
        # Drop to normal user privileges
        os.setegid(normal_gid)
        os.seteuid(normal_uid)
        yield
    finally:
        # Restore to original user and group IDs
        os.seteuid(original_euid)
        os.setegid(original_egid)


def createServiceFile(echo=False):
    # Finding the path to the 'remla' executable
    executablePath = subprocess.check_output(['which', 'remla'], text=True).strip()
    executablePath = Path(executablePath)
    if not executablePath.exists():
        raise FileNotFoundError("The 'remla' executable was not found in the expected path.")

    # Setting the PATH environment variable
    binPath = executablePath.parent  # Assuming the 'remla' binary's directory includes the necessary Python environment
    user = homeDirectory.owner()
    # Service file content
    serviceContent = f"""
[Unit]
Description=Remla
After=network.target
        
[Service]
User={user}
Group={user}
WorkingDirectory={remoteLabsDirectory}
ExecStart={executablePath} run {"-w" if echo else ""} -f
ExecStartPre=/bin/sleep 5
Restart=always
Environment="PATH={binPath}:/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin"
StandardOutput=append:/var/log/remla.log
StandardError=append:/var/log/remla.log


[Install]
WantedBy=multi-user.target
"""

    # Writing the service file
    serviceFilePath = Path('/etc/systemd/system/remla.service')
    serviceFilePath.write_text(serviceContent)
    try:
        subprocess.run(["sudo", "systemctl", "daemon-reload"], check=True)
        success(f"Service file created at {serviceFilePath}")
    except subprocess.SubprocessError:
        alert("Could not restart remla daemon.")


def cleanupPID():
    typer.echo("Cleaning up...")
    if os.path.exists(pidFilePath):
        os.remove(pidFilePath)
    sys.exit(0)

def getCallingUserID():
    sudo_uid = os.environ.get("SUDO_UID")
    if sudo_uid:
        return int(sudo_uid)
    else:
        # If SUDO_UID is not set, fall back to the current user's ID
        return os.getuid()

def bothOrNoneAssigned(x, y):
    if x is None and y is None:
        return True
    elif x is not None and y is not None:
        return True
    else:
        return False

# def runAsUser(nonPrivilegedUid=1000):
#     def decoratorRunAsUser(func):
#         @functools.wraps(func)
#         def wrapperRunAsUser(*args, **kwargs):
#             currentUid = os.geteuid()
#             try:
#                 # Switch to non-privileged user
#                 os.seteuid(nonPrivilegedUid)
#                 result = func(*args, **kwargs)
#             finally:
#                 # Always switch back to the original user
#                 os.seteuid(currentUid)
#             return result
#         return wrapperRunAsUser
#     return decoratorRunAsUser



import typer
from rich import print as rprint
from rich.prompt import IntPrompt
from rich.panel import Panel
from functools import wraps
from pathlib import Path

def warning(message: str):
    """Prints a warning message in orange with an alert emoji."""
    rprint(f"⚠️ [orange1] {message} [/orange1] ⚠️ ")

def alert(message: str):
    """Prints a alerting message in red with an alert emoji."""
    rprint(f"🚨 [red] {message} [/red] 🚨 ")

def success(message: str):
    """Prints a sucess message in green with a thumbs up emoji"""
    rprint(f"✅ [green] {message} [/green] ✅ ")

def echoResult(result, goodMessage, badMessage=None):
    if result:
        success(goodMessage)
    else:
        if badMessage is not None:
            alert(badMessage)
        raise typer.Abort()

def panelDisplay(message:str, **kwargs):
    panel  = Panel(message, expand=False, **kwargs)
    rprint(panel)

def remlaPanel(message:str):
    panelDisplay(message, title="Remla", border_style="#febc11")



import typer
from ruamel.yaml import YAML
from ruamel.yaml.nodes import ScalarNode
from pathlib import Path, PosixPath, PurePosixPath, PurePath
from typing import List, Any
from remla.labcontrol import Controllers
from collections import defaultdict


# Initialize the YAML parser
yaml = YAML(typ='safe')
# yaml.preserve_quotes = True  # Preserve quotes style
yaml.indent(mapping=2, sequence=4, offset=2)  # Set indentation, optional
yaml.default_flow_style = False
# yaml.allow_unicode = True

# Used to convert pathLib paths too yml files and vice versa.
def path_representer(dumper, data):
    return dumper.represent_scalar('!path', str(data))

def path_constructor(loader, node):
    value = loader.construct_scalar(node)
    return Path(value)

# Add the custom representer for Path objects
for cls in [Path, PosixPath, PurePosixPath, PurePath]:
    yaml.representer.add_representer(cls, path_representer)
# Add the custom constructor for Path objects
yaml.constructor.add_constructor('!path', path_constructor)


def createDevicesFromYml(deviceData:dict) -> dict[Any]:
    """
    Create and initialize devices from a YAML configuration file.

    This function reads a YAML file that defines devices and their properties,
    including dependencies on other devices. It ensures all devices are initialized
    in the correct order, even when some devices depend on others being created first.

    :param deviceDict: Dictionary of device names and properties read directly from the yaml file.
    :type deviceDict: dict
    :return: A list of initialized device objects.
    :rtype: List[Any]
    :raises ValueError: If there is a circular dependency detected among devices.
    """

    # Dictionary to hold name -> object mapping
    devices = {}

    # Set to track objects currently being initialized to detect circular dependencies
    inProgress = set()

    def resolveDependencies(deviceName):
        typer.echo(f"Resolving dependency for {deviceName}")
        """
        Recursively initialize a device and its dependencies.

        This internal function creates a device object by its name after resolving
        all necessary dependencies specified in the configuration. It supports
        recursive resolution to handle nested dependencies.

        :param deviceName: The name of the device to initialize.
        :type deviceName: str
        :return: An initialized device object.
        :rtype: Any
        :raises ValueError: If circular dependencies are detected.
        """
        # Check if device is already created and return it
        if deviceName in devices:
            return devices[deviceName]

        # Circular dependency check
        if deviceName in inProgress:
            raise ValueError(f"Circular dependency detected involving {deviceName}")

        # Mark this device as being in the process of initialization
        inProgress.add(deviceName)

        # Retrieve the class type and initialization arguments for this device
        deviceDetails = deviceData[deviceName]

        # cls = globals()[deviceDetails['type']]
        cls = getattr(Controllers, deviceDetails['type'])
        initArgs = {k: v for k, v in deviceDetails.items() if k not in ['type', 'name']}

        # Resolve dependencies for each initialization argument
        for arg, value in initArgs.items():
            if isinstance(value, str) and value in deviceData:
                initArgs[arg] = resolveDependencies(value)

        # Create the device instance and add to the devices dictionary
        device = cls(name=deviceName, **initArgs)
        devices[deviceName] = device

        # Remove device from inProgress set
        inProgress.remove(deviceName)

        return device

    # Initialize all devices by resolving their dependencies
    for name in deviceData.keys():
        if name not in devices:
            resolveDependencies(name)

    return devices



from remla.labcontrol.Controllers import gpio
import typer
from remla.systemHelpers import *
from rich.prompt import Prompt
from remla.settings import *
from remla.typerHelpers import *
from remla.systemHelpers import *
from typing_extensions import Annotated
from typing import Union
import re

app = typer.Typer(no_args_is_help=False)

addOverlayCmd = "sudo dtoverlay i2c-gpio bus=<bus> i2c_gpio_sda=<sda> i2c_gpio_scl=<scl>"
removeOverlayCmd = "sudo dtoverlay -r <index>"
armI2CCmd = "sudo dtparam i2c_arm=<state>"

@app.command()
def new(
        bus: Annotated[Union[str, None], typer.Option("--bus", "-b", help="The bus number you want." )]=None,
        sda: Annotated[Union[str,None], typer.Option("--sda", "-d", help="SDA Data Line GPIO number in BCM.")]=None,
        scl: Annotated[Union[str,None], typer.Option("--scl", "-c", help="SCL Clock Line GPIO Number in BCM")]=None,
        arm: Annotated[Union[bool,None], typer.Option("--arm", "-a", help="Turns on the builtin I2C on GPIO 2 & 3")]=False
        ):

    """
    MUST BE RUN AS ROOT. (use sudo)
    
    This command sets up a new i2c bus by writing right to the /boot/firmware/config.txt file.
    It will then try to load an overlay so that you can use it before reboot.
    
    It works in two methods you can just run `sudo remla i2c new` to get an interactive prompt for how to add a new i2c bus to the 
    system (this is the preferred method). Additionally you can use the the -b, -d, -s flags to provide 
    your own bus, sda, and scl respectively. Just put the flag followed by a number, i.e. -b 11 for bus 11

    You must provide a bus if you do it this way. You must also provide both an sda and scl pin or provide neither. If neither is provided 
    then the defaults of GPIO 23 and 24 are selected.

    Finally, you can use the -a flag to turn on the default i2c-1 bus if it is not on.
    
    Note: All pin numbers should use the BCM convention.
    """
        ####### Verify that we are running as sudo ######
    if os.geteuid() != 0:
        alert("This script must be run as root.")
        typer.echo("Try running with sudo")
        raise typer.Abort()
        # Check if running interactive mode
    if bus is None and sda is None and scl is None and not arm:
        _i2cInteractive()
    elif arm:
        config = bootConfigPath.read_text()
        arm_line = "dtparam=i2c_arm=<state>"
        i2c_arm, i2c_arm_line_no, i2c_overalys = getI2COverlays()
        if i2c_arm:
            warning("I2C arm is already in config.txt. If you don't see it when you think you should try a reboot")
            return
        if i2c_arm_line_no < 0:
            bootConfigPath.write_text(config + "\n" + arm_line.replace("<state>", "on"))
        else:
            bootConfigPath.write_text( config.replace(arm_line.replace("<state>", "off"), arm_line.replace("<state>", "on")) )
        success("Added i2c_arm to config.txt. It will now be automatically loaded on reboot")
        cmd = armI2CCmd.replace("<state>", "on")
        try:
            subprocess.run(cmd.split(), check=True)
            success("Also loaded in i2c_arm right now, so you can use it before reboot.\nTo test it out, run `i2cdetect -y 1`")
        except subprocess.CalledProcessError as e:
            warning(f"We tried to load I2C right now but something went wrong. You likely need to reboot to have access to I2C. The error message was\n\n{e}")
            
    else:
        #Check if bus is given
        try:
            int(bus)
            if bus is None or int(bus) <= 2:
                    raise ValueError
        except ValueError:
            alert("Bus must be an integer greater than or equal to 2, and probably less than 16")
            raise typer.Abort()
        except TypeError:
            alert("You must provide a bus argument with `-b` flag followed by an integer from 2-16")
            raise typer.Abort()

        # Check if SDA and SCL are both none, or provided.
        if not bothOrNoneAssigned(sda, scl):
                alert("You need to do one of the following:\n\
        1. Assign both SDA and SCL to numbers on your representing GPIO pins in BCM format\n\
        2. Not assign either SDA or SCL in which case BCM pins 23, and 24 will be used.")
                raise typer.Abort()

        sda = sda or "23"
        scl = scl or "24"
        
        if sda==scl:
                alert("SDA and SCL must be different GPIO pins")
                raise typer.Abort()
                
        i2c_arm, i2c_arm_line_no, i2c_overlays = getI2COverlays()
        forbidden_gpios = {0,1,2,3}
        preferred_gpios = {17,27,22,23,24,25,5,6,16,26}

        if {int(sda), int(scl)} & forbidden_gpios:
            alert(f"You can't use any of the GPIOs in the the following {forbidden_gpios}")
            raise typer.Abort()

        if not ({int(sda), int(scl)} & preferred_gpios):
            warning(f"If you can, try to use one of the following GPIO's as they have no other uses:\n{preferred_gpios}")
        
        conflicts = []
        for overlay in i2c_overlays:
            in_use = f""
            if overlay["bus"] == bus:
                in_use += f"BUS: {bus}\t"
                #alert(f"Bus {bus} is already in use. Consider using a different bus number or using `remla i2c update` to update that bus")
            if overlay["sda"] == sda:
                in_use += f"GPIO_SDA: {sda}\t"
            if overlay["scl"] == scl:
                in_use += f"GPIO_SCL: {scl}"
            if in_use != "":
                line = _createI2CLine(overlay['bus'], overlay['sda'], overlay['scl'])
                conflicts.append(line)

        if len(conflicts) != 0:
            message = "\nYou should run `remla i2c delete` to remove conflicting i2c conditions"
            message += "\n".join(conflicts)
            alert(message)
            raise typer.Abort()
        line = _createI2CLine(bus, sda, scl)
        with open(bootConfigPath, "a") as config:
            config.write("\n"+line)
        success(f"Updated {bootConfigPath} with new overlay:\n\t{line}\n Overlay will automatically be activated on next reboot. Loading in overlay right now as well...")
        

        cmd = addOverlayCmd.replace("<bus>", bus).replace("<sda>", sda).replace("<scl>", scl).strip()
        try:
            subprocess.run(cmd.split(), check=True)
            success(f"Added new i2c bus. You can check by running the command `i2cdetect -y {bus}`")
        except subprocess.CalledProcessError as e:
            alert(f"There was an error loading the overlay. Error read:\n {e}")

@app.command(help="The will list out busses that are avaialble and then remove you selected one from the /boot/firmware/config.txt.\
                    A restart is required for you changes to take effect.\n\
                    This can also be used to turn off the default i2c-1 bus if needed.")
def delete():
    i2c_arm, i2c_arm_line_no, i2c_overlays = getI2COverlays()
    choice_string = ""
    choices = len(i2c_overlays)
    if not i2c_arm and choices == 0:
        warning("No I2C enteries to delete. Exiting...")
        return
    for index, overlay in enumerate(i2c_overlays):
        choice_string += f"  {index+1}. {_createI2CLine(overlay['bus'], overlay['sda'], overlay['scl'])}\n"
    if i2c_arm:
        choice_string += f"  {len(i2c_overlays) + 1}. dtparam=i2c_arm=on"
        choices += 1
    deletionChoice = IntPrompt.ask(f"Which overlay do you want to delete?\n{choice_string}\n", choices=[str(i) for i in range(1, choices+1)])
    if deletionChoice != choices or not i2c_arm:
        config = bootConfigPath.read_text().splitlines()
        del config[i2c_overlays[deletionChoice-1]["line_no"]]
        bootConfigPath.write_text("\n".join(config) + "\n")
    else:
        config = bootConfigPath.read_text()
        config = config.replace("dtparam=i2c_arm=on", "dtparam=i2c_arm=off")
        bootConfigPath.write_text(config)
    success("Successfully updated config.txt.")

def getI2COverlays():
    i2c_overlays = []
    i2c_arm = False
    i2c_arm_line_no = -1
    with open(bootConfigPath, "r") as config:
        for index, line in enumerate(config):
            if line.startswith("dtoverlay=i2c-gpio"):
                i2c_overlays.append(_extractI2CInfo(line.strip()))
                i2c_overlays[-1]["line_no"] = index
            elif line.startswith("dtparam=i2c_arm="):
                if line.strip().split("=")[-1] == "on":
                    i2c_arm=True
                i2c_arm_line_no = index

    return (i2c_arm, i2c_arm_line_no, i2c_overlays)

# Function to find matches and extract information with improved flexibility
def _extractI2CInfo(line):
            # Use `findall` to capture all matching parts, regardless of order
    matches = re.findall(r"(bus=(\d+)|i2c_gpio_sda=(\d+)|i2c_gpio_scl=(\d+))", line)
    result = {"bus": None, "sda": None, "scl": None}
        
    for match in matches:
        if match[1]:  # `bus`
            result["bus"] = match[1]
        elif match[2]:  # `sda`
            result["sda"] = match[2]
        elif match[3]:  # `scl`
            result["scl"] = match[3]
        
    return result

def _createI2CLine(bus, sda, scl):
    line = f"dtoverlay=i2c-gpio"
    if bus:
        line += f",bus={bus}"
    if sda:
        line += f",i2c_gpio_sda={sda}"
    if scl:
        line += f",i2c_gpio_scl={scl}"
    return line

def _i2cInteractive():
    i2c_arm, i2c_arm_line_no, i2c_overlays = getI2COverlays()
    preferred_gpios = {"17","27","22","23","24","25","5","6","16","26"}
    other_gpios = {"4", "14", "15", "18", "9", "10", "11", "7", "8", "13", "19", "20", "21"}
    allowed_buses = {"2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15"}
    
    for overlay in i2c_overlays:
        preferred_gpios.discard(overlay["sda"])
        other_gpios.discard(overlay["sda"])
        preferred_gpios.discard(overlay["scl"])
        other_gpios.discard(overlay["scl"])
        allowed_buses.discard(overlay["bus"])
    
    allowed_gpios = list(preferred_gpios | other_gpios)
    allowed_gpios.sort(key=lambda x: int(x))
    allowed_buses = list(allowed_buses)
    allowed_buses.sort(key = lambda x: int(x))
    available_preferred_gpios = list(preferred_gpios)
    available_preferred_gpios.sort(key=lambda x: int(x))
    bus_num = Prompt.ask("What bus number do you want to use?\n", choices=allowed_buses)
    gpio_prompt = f"""Which GPIO pin do you want to use for your <type>?\n
Note: You are not required to choose from this list, but the preferred list
below are GPIO's that aren't used by any builtin software\n
Preferred List: {available_preferred_gpios}\n
"""
    print(allowed_gpios)
    sda_num = Prompt.ask(gpio_prompt.replace("<type>", "SDA"), choices=allowed_gpios)
    allowed_gpios.remove(sda_num)
    scl_num = Prompt.ask(gpio_prompt.replace("<type>", "SCL"), choices=allowed_gpios)
    
    with open(bootConfigPath, "a") as config:
        config.write("\n" + _createI2CLine(bus_num, sda_num, scl_num))

    success("Added i2c_arm to config.txt. It will now be automatically loaded on reboot")
    cmd = addOverlayCmd.replace("<bus>", bus_num).replace("<sda>", sda_num).replace("<scl>", scl_num).strip()
    try:
        subprocess.run(cmd.split(), check=True)
        success(f"Also loaded in i2c_arm right now, so you can use it before reboot.\nTo test it out, run `i2cdetect -y {bus_num}`")
    except subprocess.CalledProcessError as e:
        warning(f"We tried to load I2C right now but something went wrong. You likely need to reboot to have access to I2C. The error message was\n\n{e}")

if __name__ == "__main__":
    results = getI2COverlays()
    print(results)




from .Controllers import (
    AbsorberController,
    ArduCamMultiCamera,
    DCMotorI2C,
    ElectronicScreen,
    FilterStepperI2C,
    FS5103RContinuousMotor,
    GeneralPWMServo,
    Keithley2000Multimeter,
    Keithley6514Electrometer,
    LimitSwitch,
    Multiplexer,
    PDUOutlet,
    Plug,
    PololuDCMotor,
    PololuStepperMotor,
    PushButton,
    PWMChannel,
    S42CStepperMotor,
    SingleGPIO,
    StepperI2C,
    StepperSimple,
)
from .Experiment import Experiment



import socket
import sys
import os
import time
from signal import signal, SIGINT
import json
import logging
import threading
import queue
import RPi.GPIO as gpio
from pathlib import Path
import asyncio
import websockets
from remla.settings import *
from collections import deque

class NoDeviceError(Exception):

    def __init__(self, device_name):
        self.device_name = device_name

    def __str__(self):
        return "NoDeviceError: This experiment doesn't have a device, '{0}'".format(self.device_name)

class Experiment(object):

    def __init__(self, name, host="localhost", port=8675, admin=False):
        self.name = name
        self.host = host
        self.port = port
        self.devices = {}

        self.lockGroups = {}
        self.lockMapping = {}

        self.allStates = {}
        self.clients = deque()
        self.activeClient = None

        self.initializedStates = False
        self.admin = admin
        self.logPath = logsDirectory / f"{self.name}.log"
        # self.jsonFile = os.path.join(self.directory, self.name + ".json")
        logging.basicConfig(filename=self.logPath, level=logging.INFO,
                            format="%(levelname)s - %(asctime)s - %(filename)s - %(funcName)s \r\n %(message)s \r\n")
        logging.info("""
        ##############################################################
        ####                Starting New Log                      ####
        ##############################################################    
        """)

    def addDevice(self, device):
        device.experiment = self
        logging.info("Adding Device - " + device.name)
        self.devices[device.name] = device


    def addLockGroup(self, name:str, devices):
        lock = threading.Lock()
        self.lockGroups[name] = devices
        for deviceName in devices:
            self.lockMapping[deviceName.name] = name

    def recallState(self):
        logging.info("Recalling State")
        with open(self.jsonFile, "r") as f:
            self.allStates = json.load(f)
        for name, device in self.devices.items():
            device.setState(self.allStates[name])
        self.initializedStates = True

    def getControllerStates(self):
        logging.info("Getting Controller States")
        for name, device in self.devices.items():
            self.allStates[name] = device.getState()
        with open(self.jsonFile, "w") as f:
            json.dump(self.allStates, f)
        self.initializedStates = True


    async def handleConnection(self, websocket, path):
        print("Connection!:", websocket, path)
        self.clients.append(websocket)  # Track all clients by their WebSocket
        try:
            if self.activeClient is None and self.clients:
                self.activeClient = websocket
                await self.sendMessage(websocket, "You have control of the lab equipment.")
            else:
                await self.sendMessage(websocket, "You are connected but do not have control of the lab equipment.")
            async for command in websocket:
                if websocket == self.activeClient:
                    await self.processCommand(command, websocket)
                else:
                    await self.sendMessage(websocket, "You do not have control to send commands.")
        finally:
            if websocket == self.activeClient:
                self.activeClient = None  # Reset control if the active client disconnects
            self.clients.pop()  # Remove client from tracking

    async def processCommand(self, command, websocket):
        print(f"Processing Command {command} from {websocket}")
        logging.info("Processing Command - " + command)
        deviceName, cmd, params = command.strip().split("/")
        params = params.split(",")
        if deviceName not in self.devices:
            print("Raising no device error")
            raise NoDeviceError(deviceName)

        await self.runDeviceMethod(deviceName, cmd, params, websocket)

    async def runDeviceMethod(self, deviceName, method, params, websocket):
        device = self.devices.get(deviceName)

        lockGroup = self.lockMapping.get(deviceName)
        if lockGroup:
            print("Locak group true")
            with self.lockGroups[lockGroup]:
                print("lock group pre result")
                result = await self.runMethod(device, method, params)
                print(f"Lock group result {result}")
        else:
            print("No lockgroup")
            result = await self.runMethod(device, method, params)
            print(f"No lockgroup result: {result}")
        if result is not None:
            await self.sendMessage(websocket, f"{deviceName} - {result}")
        else:
            await self.sendMessage(websocket, f"{deviceName} ran {method}")

    async def runMethod(self, device, method, params):
        print("Running method")
        if hasattr(device, 'cmdHandler'):
            print(f"Device has cmdHandler {getattr(device, 'cmdHandler')}")
            func = getattr(device, 'cmdHandler')
            print(f"Got Commmand handler: {func}")
            loop = asyncio.get_running_loop()
            print(f"Running method {method} on the {device}")
            result = await loop.run_in_executor(None, func, method, params, device.name)
            return result
        else:
            print(f"Device {device} does not have cmdHandler method")
            logging.error(f"Device {device} does not have cmdHandler method")
            raise

    def startServer(self):
        # This function sets up and runs the WebSocket server indefinitely
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        start_server = websockets.serve(self.handleConnection, self.host, self.port)

        print(f"Server started at ws://{self.host}:{self.port}")
        loop.run_until_complete(start_server)
        loop.run_forever()

    async def sendDataToClient(self, websocket, dataStr:str):
        try:
            await websocket.send(dataStr)
        except websockets.exceptions.ConnectionClosed:
            logging.warning(f"Failed to send message: {dataStr} - Connection was closed.")
            print(f"Failed to send message: {dataStr} - Connection was closed.")
    async def sendMessage(self, websocket, message:str):
        updatedMessage = f"MESSAGE: {message}"
        await self.sendDataToClient(websocket,updatedMessage)

    async def sendAlert(self, websocket, alertMsg:str):
        updatedAlertMsg = f"ALERT: {alertMsg}"
        await self.sendDataToClient(websocket, updatedAlertMsg)

    async def sendCommandToClient(self, websocket, command:str):
        updatedCommand = f"COMMAND: {command}"
        await self.sendDataToClient(websocket, updatedCommand)


    def deviceNames(self):
        names = []
        for deviceName in self.devices:
            names.append(deviceName)
        return names

    async def onClientDisconnect(self, websocket):
        # Remove client from the client queue if they disconnect
        if websocket in self.clients:
            self.clients.remove(websocket)
        if websocket == self.activeClient:
            self.activeClient = None
            # Pass control to the next available client in the queue
            while self.clientQueue:
                potentialController = self.clientQueue.popleft()
                if potentialController.open:
                    self.activeClient = potentialController
                    await self.sendMessage(self.activeClient, "You now have control of the lab equipment.")
                    break
            if not self.activeClient:
                logging.info(f"No active clients")
                self.activeClient = None

            logging.info(f"Active client disconnected: {websocket}.")
        else:
            logging.info(f"Non-active client disconnected: {websocket}.")

    def exitHandler(self, signalReceived, frame):
        logging.info("Attempting to exit")
        if self.socket is not None:
            self.socket.close()
            logging.info("Socket is closed")

        if self.messengerSocket is not None:
            self.messengerSocket.close()
            logging.info("Messenger socket closed")

        if not self.admin:
            logging.info("Looping through devices shutting them down.")
            for deviceName, device in self.devices.items():
                logging.info("Running reset and cleanup on device " + deviceName)
                device.reset()
            logging.info("Everything shutdown properly. Exiting")
        gpio.cleanup()
        exit(0)

    def closeHandler(self):
        logging.info("Client Disconnected. Handling Close.")
        if self.connection is not None:
            self.connection.close()
            logging.info("Connection to client closed.")
        if not self.admin:
            for deviceName, device in self.devices.items():
                logging.info("Running reset on device " + deviceName)
                device.reset()

    def setup(self):
        try:
            if not self.initializedStates:
                self.getControllerStates()
            if not os.path.exists(self.socketPath):
                f = open(self.socketPath, 'w')
                f.close()

            if self.messenger is not None:
                self.messengerThread = threading.Thread(target=self.messenger.setup, daemon=True)
                self.messengerThread.start()
            os.unlink(self.socketPath)
            self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
            signal(SIGINT, self.exitHandler)
            self.socket.bind(self.socketPath)
            self.socket.listen(1)
            self.socket.setTimeout(1)
            self.__waitToConnect()
        except OSError:
            if os.path.exists(self.socketPath):
                print(f"Error accessing {self.socketPath}\nTry running 'sudo chown pi: {self.socketPath}'")
                os._exit(0)
                return
            else:
                print(f"Socket file not found. Did you configure uv4l-uvc.conf to use {self.socketPath}?")
                raise
        except socket.error as err:
            logging.error("Socket Error!", exc_info=True)
            print(f"Socket error: {err}")



import shutil

import typer
from .systemHelpers import getSettings, clearDirectory, promptForNumericFile, updateRemlaNginxConf
from rich import print as rprint
from rich.prompt import Prompt, Confirm
from remla.settings import *
from remla.typerHelpers import *
import validators
from remla.customvalidators import domainOrHostnameValidtor, portValidator
from remla.systemHelpers import *
from typing_extensions import Annotated
from remla.yaml import yaml
from typing import Optional

app = typer.Typer(no_args_is_help=False)

@app.command(name="int")
def interactive():
    """
    Runs an interactive setup that searches files and prompts users which one they want to use.
    """
    numNetworkSettings = 3
    numWebsiteSettings = 3 # Change these if more settings are required.
    remlaSettings = getSettings()
    rprint(f"Searching {remoteLabsDirectory} for files.")
    labSelection = promptForNumericFile("Which lab file do you want to use:",
                                        remoteLabsDirectory,
                                 "*.yml",
                                 f"You don't have any labs setup. Please run `remla new` or put lab yml in the {remoteLabsDirectory} directory.")

    rprint(f"Setting up {labSelection.name}")
    labSettings = yaml.load(labSelection)
    remlaSettings["currentLab"] = labSelection.relative_to(remoteLabsDirectory)
    #TODO: Add Validation Function to check it labSettings make sense.
    networkSettings = labSettings.get("network", {})
    websiteSettings = labSettings.get("website", {})
    networkUpdate = False
    if None in networkSettings.values() or len(networkSettings) < 3:
        rprint("It looks like your network isn't fully configured. Let me help.")
        networkUpdate = True
    else:
        networkUpdate = Confirm.ask("Do you want to change your network settings?")

    if networkUpdate:
        validPort = False
        while not validPort:
            port = IntPrompt.ask("Which http port do you want your website to be accessible on?", default=8080)
            validPort = portValidator(port)
        validwsPort = False
        while not validwsPort:
            wsPort = IntPrompt.ask("Which websocket port do you want to communicate on?", default=8675)
            validwsPort = portValidator(wsPort)
            if validwsPort and port == wsPort:
                validwsPort = False
                warning("You can't use the same port as the HTTP port.")
        validDomain = False
        while not validDomain:
            domain = Prompt.ask("What domain name do you want to use? e.g. example.com, 127.0.0.1, hostname")
            validDomain = domainOrHostnameValidtor(domain)

        networkSettings["port"] = port
        networkSettings["domain"] = domain
        networkSettings["wsPort"] = wsPort


    websiteUpdate = False
    if websiteSettings.get("index") is None:
        rprint("It looks like your website isn't fully configured. Let me help.")
        websiteUpdate = True
    else:
        websiteUpdate = Confirm.ask("Do you want to change your website settings?")
    if websiteUpdate:
        question = "Which file is should act as your index.html page? (The labs main webpage)"
        indexSelection = promptForNumericFile(question,
                                              remoteLabsDirectory,
                                              "*.html",
                                              f"There are no HTML files in your {remoteLabsDirectory} directory."
                                              )
        rprint(f"You have selected {indexSelection.relative_to(remoteLabsDirectory)}")
        websiteSettings["index"] = indexSelection.relative_to(remoteLabsDirectory)


    commonStaticFolderDesire = Confirm.ask("Do you have a common static folder for all your websites?", default=False)
    commonStaticUpdate = False
    if commonStaticFolderDesire:
        staticExist =  (remoteLabsDirectory/"static").exists()
        if not staticExist:
            alert(f"You don't currently have a static folder in {remoteLabsDirectory} to copy. Will not update.")
            websiteSettings["commonStaticFolder"] =  False
        else:
            websiteSettings["commonStaticFolder"] = True
            #TODO: There may be some text substitution that we need to do here, depending on lab choice.
    else:
        websiteSettings["commonStaticFolder"] = False

    staticUpdate = Confirm.ask("Do you have a specific static folder for your labs website?", default=True)
    if staticUpdate:
        question = "Which static folder is the one for the lab you want to setup?"
        staticSelection = promptForNumericFile(question,
                                              remoteLabsDirectory,
                                              "static",
                                              f"There are no static Directories in your {remoteLabsDirectory} directory."
                                              )
        rprint(f"You have selected {staticSelection.relative_to(remoteLabsDirectory)}")
        websiteSettings["staticFolder"] = staticSelection.relative_to(remoteLabsDirectory)
    elif  websiteSettings["staticFolder"] is not None and not websiteSettings["staticFolder"].exists():
        alert(f"Your provided static folder {websiteSettings['staticFolder']} doesn't exist")
        raise typer.Abort()

    labSettings["network"] = networkSettings
    labSettings["website"] = websiteSettings
    _setup(labSettings)
    yaml.dump(remlaSettings, settingsDirectory/"settings.yml")
    yaml.dump(labSettings, remoteLabsDirectory/remlaSettings["currentLab"])

def _setup(labSettings:dict)->None:
    createServiceFile(False)
    networkSettings = labSettings["network"]
    websiteSettings = labSettings["website"]
    updateRemlaNginxConf(networkSettings["port"], networkSettings["domain"], networkSettings["wsPort"])
    clearDirectory(websiteDirectory)
    shutil.copy(remoteLabsDirectory / websiteSettings["index"], websiteDirectory / "index.html")
    if websiteSettings["commonStaticFolder"]:
        shutil.copytree(remoteLabsDirectory / "static", websiteDirectory / "static", dirs_exist_ok=True)
    if websiteSettings["staticFolder"] is not None:
        shutil.copytree(remoteLabsDirectory/websiteSettings["staticFolder"], websiteDirectory / "static", dirs_exist_ok=True)
    requiredFiles = ["reader.js", "mediaMTXGetFeed.js", "remlaSocket.js"]
    for file in requiredFiles:
        shutil.copy(setupDirectory / file, websiteJSDirectory)

    shutil.copytree(websiteDirectory, nginxWebsitePath, dirs_exist_ok=True)

@app.command()
def lab(labfile: Annotated[str, typer.Argument()],
        port: Annotated[int, typer.Option()]=None,
        wsport: Annotated[int,typer.Option()]=None,
        domain: Annotated[str, typer.Option()]=None,
        cstaticon: Annotated[bool, typer.Option()]=False,
        cstaticoff: Annotated[bool,typer.Option()]=False,
        staticfolder: Annotated[str, typer.Option()]=None,
        ):
    labFileSuffix = Path(labfile).suffix

    if not labFileSuffix == ".yml":
        alert("The lab file you provided is not a yml file.")
        raise typer.Abort()
    labfile = labfile.strip("/")


    alertMsg = f"No files with name {labfile} exist in {remoteLabsDirectory}"
    potentialLabs = searchForFilePattern(remoteLabsDirectory, labfile, (alertMsg, alert))
    warnMsg = f"You have duplicate files with name {labfile} within {remoteLabsDirectory}."
    if not uniqueValidator(potentialLabs, (warnMsg,warning), processMethod=lambda x: x.name, abort=False):
        labFilePath = promptForNumericFile("Select which version you want.", remoteLabsDirectory, labfile)
    else:
        labFilePath = remoteLabsDirectory / potentialLabs[0]

    remlaSettings = getSettings()
    remlaSettings["currentLab"] = labFilePath.relative_to(remoteLabsDirectory)
    labSettings = yaml.load(labFilePath)

    # If the user doesn't provide changes to the settings when setting up the lab
    # then load them in from the settings file.
    if port is None:
        port = labSettings["network"]["port"]
    if wsport is None:
        wsPort = labSettings["network"]["wsPort"]
    else:
        wsPort = wsport
    if domain is None:
        domain = labSettings["network"]["domain"]
    if not cstaticon and not cstaticoff:
        commonStaticFolder = labSettings["website"]["commonStaticFolder"]
    elif cstaticon:
        commonStaticFolder = True
    else:
        commonStaticFolder = False
    if staticfolder is None:
        staticFolder = labSettings["website"]["staticFolder"]
    else:
        staticFolder = staticfolder

    if not portValidator(port):
        raise typer.Abort()
    if not portValidator(wsPort):
        raise typer.Abort()
    elif wsPort == port:
        alert("The http port and the websocket port can't be the same!")
        raise typer.Abort()
    if not domainOrHostnameValidtor(domain):
        raise typer.Abort()

    if labSettings["website"]["index"] is None:
        alert("You haven't send an index.html file in your settings file. Please update before continuing or run `remla setup int` for guidance through a set up.")
        raise typer.Abort()
    elif not (remoteLabsDirectory/labSettings["website"]["index"]).exists():
        alert(f"Your index file in your settings doesn't exist at {labSettings['website']['index']}. Please update before continuing or run `remla setup int` for guidance through a set up.")
        raise typer.Abort()

    labSettings["network"]["port"] = port
    labSettings["network"]["wsPort"] = wsPort
    labSettings["network"]["domain"] = domain
    labSettings["website"]["commonStaticFolder"] = commonStaticFolder
    labSettings["website"]["staticFolder"] = staticFolder

    _setup(labSettings)
    yaml.dump(labSettings, labFilePath)
    yaml.dump(remlaSettings, settingsDirectory/"settings.yml")





if __name__=="__main__":
    app()



from .typerHelpers import *
import typer
from click import style
import validators
from remla.typerHelpers import *
from typing import Callable, Any

def sensorValidator(selection: str) -> int:
    validMessage = "Please input a number between 1 and 5."
    try:
        value = int(selection)
    except ValueError:
        raise RemlaBadParameter(validMessage)

    if value-1 in range(6):
        return value-1
    else:
        raise RemlaBadParameter(validMessage)

def domainOrHostnameValidtor(domainOrHostname:str, alertUser:bool=True) -> bool:

    if validators.domain(domainOrHostname):
        return True

    if validators.hostname(domainOrHostname):
        return True
    if alertUser:
        alert("Invalid Domain, Hostname, or IP Address")
    return False

def portValidator(port:int, alertUser:bool=True)->bool:
    validPort = validators.between(port, min_val=0, max_val=65535)
    definedPort = validators.between(port, min_val=0, max_val=1023)
    if not validPort:
        if alertUser:
            alert("Ports must be in the range 0-65535")
        return False
    elif definedPort:
        warning("Ports in the range 0-1023 are typically reserved. Make sure you know what you are doing.")
    return True

def urlValidator(url:str, alertUser:bool=True)->bool:
    if validators.url(url):
        return True
    if alertUser:
        alert("You did not provide a valid URL")
    return False

def uniqueValidator(items:list, invalidMsg:tuple[str,Callable[[str],None]]|None=None,
                    processMethod:Callable[[...,Any],list]|None=None, abort:bool=True) -> bool:
    if processMethod is not None:
        items = list(map(processMethod, items))
    itemSet = set(items)
    if len(itemSet) != len(items):
        if invalidMsg is not None:
            msgType = invalidMsg[1]
            msg = invalidMsg[0]
            msgType(msg)
        if abort:
            raise typer.Abort()
        return False
    return True

class RemlaBadParameter(typer.BadParameter):

    def __init__(self, message, *args, **kwargs):
        super().__init__(style(f"⚠️ {message} ⚠️ ", fg="yellow"), *args, **kwargs)




import inspect
import os
import subprocess
import sys
import time
from abc import ABC, ABCMeta, abstractmethod
from warnings import warn

import dlipower
import pigpio
import pyvisa as visa
import RPi.GPIO as gpio
import RPistepper as stp
import tplink_smarthome as tp
import typer
from adafruit_motor import stepper
from adafruit_motorkit import MotorKit

pi = pigpio.pi()
gpio.setmode(gpio.BCM)
visaManager = visa.ResourceManager("@py")


class InitialAttributeTracker:
    def __init__(self, name=None):
        self.name = name

    def __get__(self, obj, objtype=None):
        print("Getting Called")
        return obj.__dict__.get(self.name)

    def __set__(self, obj, value):
        obj.__dict__[self.name] = value
        obj.initParameters[self.name] = value

    def __delete__(self, obj):
        obj.__dict__.pop(self.name, None)
        obj.initParameters.pop(self.name, None)


class AutoTrackMeta(type):
    def __new__(cls, name, bases, dct):
        # Create the class using super().
        new_cls = super().__new__(cls, name, bases, dct)
        # Adding AttributeTracker to all initial attributes specified in __init__
        init_func = dct.get("__init__")
        if init_func is None:
            # Search in base classes
            for base in bases:
                if "__init__" in base.__dict__:
                    init_func = base.__dict__["__init__"]
                    break
        if init_func:
            # Using signature to extract argument names
            sig = inspect.signature(init_func)
            for param in sig.parameters.values():
                if param.name == "self" or param.default is not inspect.Parameter.empty:
                    continue
                # Set the descriptor for handling attribute
                setattr(new_cls, param.name, InitialAttributeTracker(param.name))
        return new_cls


class CombinedMetaClass(AutoTrackMeta, ABCMeta):
    pass


class BaseController(ABC, metaclass=CombinedMetaClass):
    def __init__(self, name):
        self.initParameters = {}
        self.name = name
        self.experiment = None
        self.state = {}
        typer.echo(f"Initialized {name} base controller")

    @property
    @abstractmethod
    def deviceType(self):
        """Subclasses must have a 'deviceType' class attribute"""
        pass

    def cmdHandler(self, cmd, params, deviceName):
        # Make the parser name, it should follow the naming convention <cmd>_parser. If there is no parser return None.
        parser_name = f"{cmd}_parser"
        parser = getattr(self, parser_name, None)

        # If parser exists, use it to parse the params.
        if parser is not None:
            params = parser(params)
        # If there is no parser print this statement for user.
        else:
            print("No Parser Found. Will just pass params to command.")

        # Now get the command method. If there isn't a method, it should throw an AttributeError.
        print("Command received by base sending off")
        try:
            method = getattr(self, cmd)
            print(method)
            if callable(method):
                response = method(params)
                return response
        except:
            print(f"{self.__class__.__name__} does not have <{cmd}> cmd")

    @abstractmethod
    def reset(self):
        pass

    def getState(self):
        return self.state

    def setState(self, state):
        self.state = state


class PDUOutlet(dlipower.PowerSwitch, BaseController):
    deviceType = "controller"

    def __init__(
        self,
        name,
        hostname,
        userid,
        password,
        timeout=None,
        outlets=[1, 2, 3, 4, 5, 6, 7, 8],
        outletMap={},
    ):
        BaseController.__init__(self, name)
        dlipower.PowerSwitch.__init__(
            self, hostname=hostname, userid=userid, password=password, timeout=timeout
        )
        self.experiment = None
        self.state = {
            1: "Off",
            2: "Off",
            3: "Off",
            4: "Off",
            5: "Off",
            6: "Off",
            7: "Off",
            8: "Off",
        }
        self.outlets = outlets
        self.outletMap = outletMap
        # self.login()

    def on(self, outletNumber):
        super().on(outletNumber)
        self.state[outletNumber] = "On"

    def off(self, outletNumber):
        super().off(outletNumber)
        self.state[outletNumber] = "Off"

    def on_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "on")
        try:
            outlet = int(params[0])
        except ValueError:
            outlet_name = params[0]
            if outlet_name not in self.outletMap:
                raise ArgumentError(outlet_name, self.outletMap, self.name, "On")
            outlet = self.outletMap[outlet_name]

        if outlet not in self.outlets:
            raise ArgumentError(
                self.name,
                "On",
                "Outlet " + params[0],
                allowed="Outlets " + str(self.outlets),
            )
        return outlet

    def off_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "off")
        try:
            outlet = int(params[0])
        except ValueError:
            outlet_name = params[0]
            if outlet_name not in self.outletMap:
                raise ArgumentError(outlet_name, self.outletMap, self.name, "Off")
            outlet = self.outletMap[outlet_name]

        if outlet not in self.outlets:
            raise ArgumentError(
                self.name,
                "Off",
                "Outlet " + params[0],
                allowed="Outlets " + str(self.outlets),
            )
        return outlet

    def reset(self):
        for outlet in self.outlets:
            self.off(outlet)


class Plug(tp.TPLinkSmartDevice, BaseController):
    deviceType = "controller"

    def __init__(self, name, host, port=9999, timeout=10, connect=True):
        print(host, port, timeout, connect)
        super().__init__(host=host, port=port, connect=connect)
        super().__init__(name)
        self.state = {"relayState": "OFF"}

    def setRelay(self, newRelay):
        print(newRelay)
        if newRelay == "OFF":
            super().send({"system": {"set_relay_state": {"state": 0}}})
        elif newRelay == "ON":
            super().send({"system": {"set_relay_state": {"state": 1}}})
        self.state["relayState"] = newRelay

    def setRelay_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "setRelay")
        return params[0]

    # def cleanup(self):
    #     super().close()

    def reset(self):
        super().send({"system": {"set_relay_state": {"state": 0}}})
        super().close()


class StepperSimple(stp.Motor, BaseController):
    deviceType = "controller"

    def __init__(self, name, pins, delay=0.02, refPoints={}):
        super().__init__(pins, delay)
        super().__init__(name)
        self.refPoints = refPoints
        self.currentPosition = 0
        self.state = {"position": self.currentPosition}

    def move(self, steps):
        print(steps)
        super().move(steps)
        super().release()
        self.currentPosition += steps
        self.state["position"] = self.currentPosition
        # self.device.release()
        self.release()

    def move_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "move")
        return int(params[0])

    def goto(self, position):
        print(position)
        endPoint = self.refPoints[position]
        self.move(endPoint - self.currentPosition)

    def goto_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "goto")
        return params[0]

    # def cleanup(self):
    #     super().cleanup()

    def reset(self):
        super().reset()


class DCMotorI2C(MotorKit, BaseController):
    deviceType = "controller"

    def __init__(self, name, terminal):
        if terminal > 4:
            self.address = 0x61
        else:
            self.address = 0x60
        super().__init__(address=self.address)
        super().__init__(name)

        self.terminal_options = {
            1: super().motor1,
            2: super().motor2,
            3: super().motor3,
            4: super().motor4,
            5: super().motor1,
            6: super().motor2,
            7: super().motor3,
            8: super().motor4,
        }
        self.device = self.terminal_options[terminal]
        self.currentPosition = 0
        self.state = {"position": self.currentPosition}

    def setup(self, style):
        pass

    def reset(self):
        pass

    def throttle(self, speed):
        self.device.throttle = speed

    def throttle_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "move")
        print("Throttle is set to: {0}".format(params[0]))
        return float(params[0])


# Initialise the first hat on the default address
# lowerBoard = MotorKit()
# Initialise the second hat on a different address
# upperBoard = MotorKit(address=0x61)


class StepperI2C(MotorKit, BaseController):
    deviceType = "controller"

    def __init__(
        self,
        name,
        terminal,
        bounds,
        delay=0.02,
        refPoints={},
        style="SINGLE",
        microsteps=8,
        limitSwitches=[],
        homeSwitch=None,
        degPerStep=1.8,
        gearRatio=1,
    ):
        if terminal > 2:
            self.address = 0x61
        else:
            self.address = 0x60
        # MotorKit.__init__(self,address=self.address, steppers_microsteps=microsteps)
        BaseController.__init__(self, name)

        # self.terminal_options = {1: super().stepper1, 2: super().stepper2, 3:super().stepper1, 4:super().stepper2}
        self.refPoints = refPoints
        self.currentPosition = 0
        # self.device = self.terminal_options[terminal]
        self.delay = delay
        self.lowerBound = bounds[0]
        self.upperBound = bounds[1]
        self.styles = {
            "SINGLE": stepper.SINGLE,
            "DOUBLE": stepper.DOUBLE,
            "MICROSTEP": stepper.MICROSTEP,
            "INTERLEAVE": stepper.INTERLEAVE,
        }
        self.style = self.styles[style]

        self.state = {"position": self.currentPosition}
        self.limitSwitches = limitSwitches
        self.homeSwitch = homeSwitch
        self.homing = False
        self.degPerStep = degPerStep
        self.gearRatio = gearRatio
        time.sleep(0.2)  # Adding this to see if it released prooperly
        # self.device.release()
        time.sleep(0.2)  # Adding this to see if it released prooperly

    def setup(self, style):
        pass

    def move(self, steps):
        print(steps)
        if steps >= 0:
            direction = stepper.BACKWARD
        else:
            direction = stepper.FORWARD
        if self.currentPosition + steps < self.lowerBound and steps < 0:
            steps = self.lowerBound - self.currentPosition
        elif self.currentPosition + steps > self.upperBound and steps > 0:
            steps = self.upperBound - self.currentPosition

        for i in range(abs(steps)):
            if len(self.limitSwitches) != 0:
                for switch in self.limitSwitches:
                    status = switch.getStatus(1)
                    if status == gpio.HIGH:
                        response = switch.switchAction(self, steps - i)
                        if response is None:
                            return "{0}/{1}/{2}".format(self.name, "position", "limit")
                        else:
                            return response

            if self.homing:
                homeStatus = self.homeSwitch.getStatus(1)
                if homeStatus == gpio.HIGH:
                    return True

            self.device.onestep(style=self.style, direction=direction)
            time.sleep(self.delay)

        self.currentPosition += steps
        self.state["position"] = self.currentPosition
        self.device.release()
        if (
            self.currentPosition == self.upperBound
            or self.currentPosition == self.lowerBound
        ):
            return "{0}/{1}/{2}".format(self.name, "position", "limit")
        else:
            return "{0}/{1}/{2}".format(self.name, "position", self.currentPosition)

    def move_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "move")
        return int(params[0])

    def adminMove(self, steps):
        if steps >= 0:
            direction = stepper.BACKWARD
        else:
            direction = stepper.FORWARD
        if self.currentPosition + steps < self.lowerBound and steps < 0:
            steps = self.lowerBound - self.currentPosition
        elif self.currentPosition + steps > self.upperBound and steps > 0:
            steps = self.upperBound - self.currentPosition
        for i in range(abs(steps)):
            self.device.onestep(style=self.style, direction=direction)
            time.sleep(self.delay)
        self.currentPosition += steps
        self.state["position"] = self.currentPosition
        self.device.release()

    def goto(self, position):
        print(position)
        endPoint = self.refPoints[position]
        response = self.move(endPoint - self.currentPosition)
        return response

    def admingoto(self, position):
        print(position)
        endPoint = self.refPoints[position]
        response = self.adminMove(endPoint - self.currentPosition)
        return response

    def goto_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "goto")
        return params[0]

    def admingoto_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "admingoto")
        return params[0]

    def degMove_parser(self, params):
        try:
            steps = float(params[0])
        except ValueError:
            raise ArgumentError(self.name, "degMove", params)
        return steps

    def degMove(self, deg):
        print("{0} degrees".format(deg))
        step = deg / self.degPerStep
        step = int(step * self.gearRatio)
        step = round(step)
        response = self.move(step)
        return response

    def homeMove(self, stepLimit=5000, additionalSteps=10):
        if self.currentPosition > 0:
            direction = -1
        elif self.currentPosition < 0:
            direction = 1
        else:
            return

        self.move(direction * stepLimit)
        self.move(direction * additionalSteps)

    def reset(self):
        if self.homeSwitch is not None:
            self.home(1)
        else:
            self.move(-self.currentPosition)

    def home(self, params):
        if self.homeSwitch is not None:
            self.homing = True
            self.customHome(self)
            self.homing = False
        else:
            print("No homing switch is attached to this motor")

    def customHome(self, motor):
        pass


"""
cyclic stepper adaptation for the StepperI2C, improved goto logic so that it
always takes the shortest path. Resetting logic is maintained to a maximum
of 1 full revolution.
"""


# TODO: Turn this into either a new class called Periodic I2C that will just generally take the shortest path or add a periodic flag to stepper I2C.
class FilterStepperI2C(StepperI2C):
    def __init__(
        self,
        name,
        terminal,
        delay=0.02,
        refPoints={},
        style="SINGLE",
        microsteps=8,
        limitSwitches=[],
        homeSwitch=None,
        degPerStep=1.8,
        gearRatio=1,
    ):
        super().__init__(
            name,
            terminal,
            (-1e9, 1e9),
            delay,
            refPoints,
            style,
            microsteps,
            limitSwitches,
            homeSwitch,
            degPerStep,
            gearRatio,
        )
        self.refPointsRaw = sorted(list(refPoints.values()))

    def goto(self, position):
        if self.currentPosition in self.refPointsRaw:
            curIdx = self.refPointsRaw.index(self.currentPosition)
            tgtIdx = self.refPointsRaw.index(self.refPoints[position])
            if abs(tgtIdx - curIdx) <= len(self.refPointsRaw) / 2:
                super().goto(position)
            else:
                # reverse direction
                if tgtIdx > curIdx:
                    dif = -(curIdx + len(self.refPointsRaw) - tgtIdx)
                # foward direction
                elif tgtIdx < curIdx:
                    dif = tgtIdx + len(self.refPointsRaw) - curIdx
                self.move(dif * (self.refPointsRaw[1] - self.refPointsRaw[0]))
                # maintain a reasonable amount of steps for resetting
                if self.currentPosition > self.refPointsRaw[-1]:
                    self.currentPosition = self.currentPosition - (
                        self.refPointsRaw[1] - self.refPointsRaw[0]
                    ) * len(self.refPointsRaw)
                if self.currentPosition < 0:
                    self.currentPosition = self.currentPosition + (
                        self.refPointsRaw[1] - self.refPointsRaw[0]
                    ) * len(self.refPointsRaw)
        else:
            super().goto(position)


class AbsorberController(BaseController):  # Removed MotorKit subclass @ZakEspley
    deviceType = "controller"

    def __init__(
        self,
        name,
        stepper,
        actuator,
        magnet,
        initialState,
        holderMap,
        downtime=0.5,
        uptime=1,
        throttle=1.0,
    ):
        super().__init__(name)
        self.downtime = downtime
        self.uptime = uptime
        self.stepper = stepper
        self.actuator = actuator
        self.magnet = magnet
        self.initialState = initialState
        self.holderMap = holderMap
        self.throttle = throttle
        self.state = {
            "loaded": {
                "s0": False,
                "s1": False,
                "s2": False,
                "s3": False,
                "s4": False,
                "s5": False,
            },
            "total": self.initialState,
        }

    def setup(self, style):
        pass

    def reset(self):
        emptyCounter = [
            ("s0", ""),
            ("s1", ""),
            ("s2", ""),
            ("s3", ""),
            ("s4", ""),
            ("s5", ""),
        ]
        movesList = self.__makeMovesList(emptyCounter)
        self.place(movesList)
        self.stepper.reset()

    def __transfer(self, slot1, slot2):
        # TODO: magnet control needs to be fixed

        absorber = self.state["total"][slot1]
        # print(f"Transfer {absorber} from {slot1} --> {slot2}")
        if self.state["total"][slot2] != "":
            print(
                f'!!!!!!!!!!!WARNING:  Slot {slot2} already full with {self.state["total"][slot2]}'
            )
            raise
        # global x
        # x += 1
        self.stepper.goto(slot1)
        self.actuator.throttle(self.actuator.throttle_parser([self.throttle]))
        time.sleep(self.downtime)
        self.actuator.throttle(self.actuator.throttle_parser([0]))
        # absorber gets attached to the magnet
        time.sleep(0.5)
        self.actuator.throttle(self.actuator.throttle_parser([-self.throttle]))
        time.sleep(self.uptime)
        self.actuator.throttle(self.actuator.throttle_parser([0]))

        self.stepper.goto(slot2)
        self.actuator.throttle(self.actuator.throttle_parser([self.throttle]))
        time.sleep(self.downtime)
        self.actuator.throttle(self.actuator.throttle_parser([0]))

        # push magnet and drop the absorber
        self.magnet.goto(45)

        self.actuator.throttle(self.actuator.throttle_parser([-self.throttle]))
        time.sleep(self.uptime)
        self.actuator.throttle(self.actuator.throttle_parser([0]))
        self.actuator.disable()
        self.magnet.goto(-90)
        time.sleep(1)
        self.magnet.disable()

        self.state["total"][slot1] = ""
        self.state["total"][slot2] = absorber
        if slot1 in self.state["loaded"]:
            self.state["loaded"][slot1] = False
        if slot2 in self.state["loaded"]:
            self.state["loaded"][slot2] = True

    def __getSlot(self, absorber):
        # print(f"GET SLOT: {absorber}")
        if absorber == "":
            return -1
        for holderSlot, ab in self.state["total"].items():
            if ab == absorber:
                return holderSlot

    def __getAbsorber(self, slot):
        return self.state["total"][slot]

    def __makeMovesList(self, newPositions):
        moveList = {"load": [], "unload": [], "internal": [], "chains": []}

        absorbers = [item[1] for item in newPositions if item[1] != ""]
        for slot, ab in newPositions:
            # slot is one of the counter slots. We will loop through all slots in the coutner.
            # ab is the absorber we want to end up in slot.

            # Get the absorber that is currently in the slot
            currentAbsInSlot = self.__getAbsorber(slot)
            # Get the location of the current absorber
            currentAbsLocation = self.__getSlot(ab)
            print(f"Absorber: {ab}  Location: {currentAbsLocation}")

            # Determine if the absorber is used later down the line
            absorberUsed = currentAbsInSlot in absorbers
            if ab == currentAbsInSlot:
                ## If the absorber is already in the correct slot there is nothing todo.
                continue

            elif ab == "" and self.state["loaded"][slot]:
                # If we want the slot to be empty, it is current loaded and the
                # loaded absorber is not used later, then unload it.
                if not absorberUsed:
                    moveList["unload"].append((slot, self.holderMap[currentAbsInSlot]))

            elif ab != "" and self.state["loaded"][slot]:
                # If we want the slot to be full and it is already full with a
                # different absorber.
                if not absorberUsed:
                    # If the currently loaded absorber is not needed, unload it.
                    # If it is used later, we will handle that one later in the loop
                    moveList["unload"].append((slot, self.holderMap[currentAbsInSlot]))
                if currentAbsLocation[0] == "s":
                    # If the absorbr we want is already in the counter
                    # add it to an internal movement
                    moveList["internal"].append((currentAbsLocation, slot))
                else:
                    # If it isn't already in the counter it must be in the holder
                    # so add it to the load movement list.
                    moveList["load"].append((currentAbsLocation, slot))

            elif ab != "" and not self.state["loaded"][slot]:
                # If we want to fill the slot and nothing is there already just move
                # if following the same rules above without an unload first.
                if currentAbsLocation[0] == "s":
                    moveList["internal"].append((currentAbsLocation, slot))
                else:
                    moveList["load"].append((currentAbsLocation, slot))

        ## Now we should identify chains.
        chains = self.__chainDetect(moveList["internal"])
        internalStarts = [item[0] for item in moveList["internal"]]
        print("Pre-chains: {0}".format(moveList))
        for chain in chains:
            for move in chain:
                moveList["internal"].remove(move)

        moveList["chains"] = chains
        print("Post-chains: {0}".format(moveList))
        return moveList

    def __chainDetect(self, movements):
        chains = []
        checked = []
        internalStarts = [item[0] for item in movements]
        internalFinish = [item[1] for item in movements]
        intersection = [slot for slot in internalStarts if slot in internalFinish]

        for move in movements:
            temp = []
            if move[0] in checked:
                continue
            # checked.append(move[0])
            if move[0] in intersection and move[1] in intersection:
                start = move[0]
                checked.append(start)
                nextSlot = move[1]
                temp.append(move)
                while nextSlot != start:
                    i = 0
                    for nextMove in movements:
                        if nextMove[0] == nextSlot:
                            checked.append(nextMove[0])
                            nextSlot = nextMove[1]
                            temp.append(nextMove)
                            i -= 1
                            break
                        # print(f"{nextMove} and {nextSlot}")
                    i += 1
                    if i > 0:
                        break
                if i <= 0:
                    chains.append(temp)
        # print(f"CHAINS: {chains}")
        return chains

    def __chaseInternal(self, internals, startingMove):
        moves = [startingMove]
        internalStarts = [item[0] for item in internals]
        # internalStartsTemp = internalStarts.copy()
        nextSlot = startingMove[1]
        while nextSlot in internalStarts:
            index = internalStarts.index(nextSlot)
            iMove = internals.pop(index)
            internalStarts.pop(index)
            moves.insert(0, iMove)
            nextSlot = iMove[1]
        return moves

    def __handleChains(self, chains, currentMoves):
        currentFinish = [item[1] for item in currentMoves]
        slots = ["s0", "s1", "s2", "s3", "s4", "s5"]
        for chain in chains:
            move1 = chain.pop(0)
            ab = self.__getAbsorber(move1[0])
            home = self.holderMap[ab]
            # for slot in slots:
            #     if slot not in currentFinish:
            #         home = slot
            #         break
            # print(f"CHAIN HOME: ({move1[0]}, {home})")
            currentMoves.append((move1[0], home))
            chain.reverse()
            for move in chain:
                currentMoves.append(move)
                # print(f"CHAIN MOVE REVERSE: {move}")
            currentMoves.append((home, move1[1]))
            # print(f"CHAIN BACK: ({home}, {move1[1]})")
            currentFinish = [item[1] for item in currentMoves]
        return currentMoves

    def __handleInternal(self, internals):
        moves = []
        if len(internals) == 0:
            return moves

        internalGroups = []
        while len(internals) != 0:
            start = internals.pop(0)
            temp = self.__chaseInternal(internals, start)
            internalGroups.append(temp)

        for group in internalGroups:
            moves += group

        return moves

    def __buildGroups(self, moveList):
        unloads = moveList["unload"]
        loads = moveList["load"]
        internals = moveList["internal"]
        print("INTERNALS:{0}".format(internals))
        unloadStarts = [item[0] for item in unloads]
        UILGroups = []
        ILGroups = []
        LGroups = []
        UIGroups = []
        IGroups = []
        moves = []

        for load in loads:
            temp = self.__chaseInternal(internals, load)
            internalStarts = [item[0] for item in internals]

            if len(temp) == 1:
                LGroups.append(temp)

            else:
                nextSlot = temp[0][1]
                # print(f"Checking for conflict with unload. Slot {nextSlot}")
                if nextSlot in unloadStarts:
                    # print("Internal Movement Conflict with Following Unload")
                    index = unloadStarts.index(nextSlot)
                    uMove = unloads.pop(index)
                    unloadStarts.remove(uMove[0])
                    temp.insert(0, uMove)
                    UILGroups.append(temp)
                else:
                    ILGroups.append(temp)

        if len(internals) > 0:
            print("Starting Internals")
            internalStarts = [item[0] for item in internals]
            internalFinish = [item[1] for item in internals]
            intersection = [item for item in internalFinish if item in internalStarts]
            print("Intersections: {0}".format(intersection))
            starts = [internal for internal in internals if internal[1] in intersection]
            for start in starts:
                internals.remove(start)
            print("INTERNALS2:{0}".format(internals))
            print("STARTS:{0}".format(starts))
            for startingI in starts:
                temp = self.__chaseInternal(internals, startingI)
                print("CHASEINTERNALS:{0}".format(temp))
                internalStarts = [item[0] for item in internals]
                nextSlot = temp[0][1]
                # print(f"Checking for conflict with unload. Slot {nextSlot}")
                if nextSlot in unloadStarts:
                    # print("Internal Movement Conflict with Following Unload")
                    index = unloadStarts.index(nextSlot)
                    uMove = unloads.pop(index)
                    unloadStarts.remove(uMove[0])
                    temp.insert(0, uMove)
                    UIGroups.append(temp)
                else:
                    IGroups.append(temp)
            for internal in internals:
                IGroups.append([internal])

        for uil in UILGroups:
            uMove = [uil.pop(0)]
            lMove = [uil.pop(-1)]
            igroup = uil

            if len(unloads) > 0:
                uTemp = unloads.pop(0)
                igroup.append(uTemp)

            elif len(UIGroups) > 0:
                uiTemp = UIGroups.pop(0)
                uTempMove = uiTemp.pop(0)
                igroup.append(uTempMove)
                lMove = lMove + uiTemp

            if len(LGroups) > 0:
                lTemp = LGroups.pop(0)
                igroup.insert(0, lTemp[0])

            elif len(ILGroups) > 0:
                ilTemp = ILGroups.pop(0)
                lTempMove = ilTemp.pop(-1)
                igroup.insert(0, lTempMove)
                uMove = ilTemp + uMove

            # moves.append(uMove + igroup + lMove)
            moves += uMove + igroup + lMove

        for il in ILGroups:
            uMove = []
            lMove = [il.pop(-1)]
            igroup = il

            if len(unloads) > 0:
                uTemp = unloads.pop(0)
                igroup.append(uTemp)

            elif len(UIGroups) > 0:
                uiTemp = UIGroups.pop(0)
                uTempMove = uiTemp.pop(0)
                igroup.append(uTempMove)
                lMove = lMove + uiTemp

            # moves.append(igroup + lMove)
            moves += igroup + lMove

        for l in LGroups:
            uMove = []
            lMove = l
            igroup = []

            if len(unloads) > 0:
                uTemp = unloads.pop(0)
                igroup.append(uTemp)

            elif len(UIGroups) > 0:
                uiTemp = UIGroups.pop(0)
                uTempMove = uiTemp.pop(0)
                igroup.append(uTempMove)
                lMove = lMove + uiTemp

            # moves.append(igroup + lMove)
            moves += igroup + lMove

        while len(unloads) > 0 or len(UIGroups) > 0:
            if len(UIGroups) > 0:
                moves = UIGroups.pop(0) + moves
            elif len(unloads) > 0:
                print(unloads)
                moves.insert(0, unloads.pop(0))

        while len(IGroups) > 0:
            moves += IGroups.pop(0)

        return moves

    def place(self, moveList):
        moves = []
        internalStarts = [item[0] for item in moveList["internal"]]
        internalFinish = [item[1] for item in moveList["internal"]]
        unloadStarts = [item[0] for item in moveList["unload"]]

        # print("\r\n ### Pairing Checks")
        uMovesTemp = moveList["unload"].copy()
        lMovesTemp = moveList["load"].copy()
        for uMove in moveList["unload"]:
            # print(f"PAIRING CHECK: {uMove}, {moveList['unload']}, {moveList['load']},")
            for lMove in moveList["load"]:
                # print(f"PARING CHECK2: {lMove}")
                if uMove[0] == lMove[1]:
                    moves.append(uMove)
                    moves.append(lMove)
                    uMovesTemp.remove(uMove)
                    lMovesTemp.remove(lMove)
                    # print(f"PARING: {uMove} with {lMove}")

        moveList["unload"] = uMovesTemp
        moveList["load"] = lMovesTemp
        unloadStarts = [item[0] for item in moveList["unload"]]

        # print("## Pairings Complete")
        # pp.pprint(moveList)

        # print("## ORGANIZING EVERYTING BUT CHAINS")
        moves += self.__buildGroups(moveList)

        if len(moveList["chains"]) > 0:
            # print("## Handling Chains")
            moves = self.__handleChains(moveList["chains"], moves)

        # pp.pprint(moveList)
        print("Moves:{0}".format(moves))
        for move in moves:
            self.__transfer(move[0], move[1])

    def place_parser(self, params):
        # if len(params) != 1:
        #     raise ArgumentNumberError(len(params), 1, "place")
        # print("Throttle is set to: {0}".format(params[0]))
        # print(params)
        temp = [item.replace("(", "").replace(")", "") for item in params]
        absorberList = []
        i = 0
        while i < len(temp):
            absorberList.append((temp[i], temp[i + 1]))
            i += 2

        moveList = self.__makeMovesList(absorberList)

        return moveList


class Multiplexer(BaseController):
    deviceType = "controller"

    def __init__(
        self,
        name,
        pins,
        inhibitorPin,
        channels=[0, 1, 2, 3, 4, 5, 6, 7],
        defaultChannel=None,
        defaultState=gpio.HIGH,
        delay=0.1,
    ):
        self.state = {}
        super().__init__(name)
        self.delay = delay
        self.pins = pins
        self.channels = channels
        self.defaultState = defaultState
        self.defaultChannel = defaultChannel
        self.inhibitorPin = inhibitorPin

        gpio.setup(self.pins, gpio.OUT)
        gpio.setup(self.inhibitorPin, gpio.OUT)

        if self.defaultChannel is not None:
            self.__setChannel(self.defaultChannel)
            gpio.output(self.inhibitorPin, self.defaultState)
        else:
            gpio.output(self.inhibitorPin, gpio.HIGH)

    def __setChannel(self, channel):
        binary = "{0:0{1}b}".format(channel, len(self.pins))
        gpio.output(self.inhibitorPin, gpio.HIGH)
        for i in range(len(self.pins)):
            pin = self.pins[i]
            state = bool(int(binary[i]))
            pinstate = gpio.LOW
            if state:
                pinstate = gpio.HIGH

            gpio.output(pin, pinstate)

    def press(self, channel):
        self.__setChannel(channel)
        gpio.output(self.inhibitorPin, gpio.LOW)
        time.sleep(self.delay)
        gpio.output(self.inhibitorPin, gpio.HIGH)

    def press_parser(self, params):
        channel = int(params[0])
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "press")
        if channel not in self.channels:
            raise ArgumentError(self.name, "press", channel, self.channels)
        return channel

    def reset(self):
        if self.defaultChannel is not None:
            self.__setChannel(self.defaultChannel)
        gpio.output(self.inhibitorPin, self.defaultState)


class Keithley6514Electrometer(BaseController):
    deviceType = "measurement"

    def __init__(self, name, usbAddress):
        super().__init__(name)
        self.state = {"setting": ""}
        self.usbAddress = usbAddress
        self.inst = visaManager.open_resource(
            f"ASRL/dev/ttyUSB{self.usbAddress}::INSTR", baud_rate=19200
        )
        self.inst.read_termination("\r\n")
        self.inst.write_termination("\r\n")

    def press(self, params):
        self.inst.write("SYST:REM")
        self.inst.write(params)
        if params != "SYST:KEY 1":
            self.inst.write("SYST:LOC")

    def press_parser(self, params):
        print(">>", params)
        return params[0]

    def reset(self):
        pass


class Keithley2000Multimeter(BaseController):
    deviceType = "measurement"

    def __init__(self, name, usbAddress):
        super().__init__(name)
        self.state = {"setting": ""}
        self.usbAddress = usbAddress
        self.inst = visaManager.open_resource(
            f"ASRL/dev/ttyUSB{self.usbAddress}::INSTR", baud_rate=19200
        )
        self.inst.read_termination("\r\n")
        self.inst.write_termination("\r\n")

    def press(self, params):
        self.inst.write("SYST:REM")
        self.inst.write(params)

    def press_parser(self, params):
        print(">>", params)
        return params[0]

    def reset(self):
        pass


class PololuStepperMotor(BaseController):
    deviceType = "controller"

    def __init__(
        self,
        name,
        stepPin,
        directionPin,
        enablePin,
        bounds,
        delay=5000,
        refPoints={},
        limitSwitches=[],
        homeSwitch=None,
        degPerStep=1.8,
        gearRatio=1,
    ):
        super().__init__(name)
        self.stepPin = stepPin
        self.directionPin = directionPin
        self.enablePin = enablePin

        pi.set_mode(self.stepPin, pigpio.OUTPUT)
        pi.set_pull_up_down(self.stepPin, pigpio.PUD_DOWN)
        pi.set_mode(self.enablePin, pigpio.OUTPUT)
        pi.set_pull_up_down(self.enablePin, pigpio.PUD_DOWN)
        pi.set_mode(self.directionPin, pigpio.OUTPUT)
        pi.set_pull_up_down(self.directionPin, pigpio.PUD_DOWN)

        pi.write(self.enablePin, 0)

        self.refPoints = refPoints
        self.currentPosition = 0
        self.state = {"position": self.currentPosition}
        self.delay = delay
        self.lowerBound = bounds[0]
        self.upperBound = bounds[1]
        # self.styles = {
        #     "SINGLE": stepper.SINGLE,
        #     "DOUBLE": stepper.DOUBLE,
        #     "MICROSTEP": stepper.MICROSTEP,
        #     "INTERLEAVE": stepper.INTERLEAVE
        # }
        # self.style = self.styles[style]

        self.state = {"position": self.currentPosition}
        self.limitSwitches = limitSwitches
        self.homeSwitch = homeSwitch
        self.homing = False
        self.degPerStep = degPerStep
        self.gearRatio = gearRatio

    def move(self, steps):
        # Choose direction
        if steps > 0:
            pi.write(self.directionPin, 1)
        else:
            pi.write(self.directionPin, 0)

        if self.currentPosition + steps < self.lowerBound and steps < 0:
            steps = self.lowerBound - self.currentPosition
        elif self.currentPosition + steps > self.upperBound and steps > 0:
            steps = self.upperBound - self.currentPosition

        # Create a train of pulses separated by delay
        pOn = pigpio.pulse(1 << self.stepPin, 0, self.delay // 2)
        pOff = pigpio.pulse(0, 1 << self.stepPin, self.delay // 2)
        pulse = [pOn, pOff]
        pi.wave_clear()
        pi.wave_add_generic(pulse)
        stepWave = pi.wave_create()

        absteps = abs(steps)
        step_y = absteps // 256
        step_x = absteps % 256

        pi.write(self.enablePin, 1)
        ## Wave Chains seems incredibly stupid.
        ## To see how to do it check out http://abyz.me.uk/rpi/pigpio/python.html#wave_chain
        pi.wave_chain(
            [
                255,
                0,  # Starts a loop
                stepWave,  # What wave to send
                255,
                1,
                step_x,
                step_y,  # Repeat loop x + 256*y times.
            ]
        )

        while pi.wave_tx_busy():
            time.sleep(0.1)

        pi.write(self.enablePin, 0)

        self.currentPosition += steps
        self.state["position"] = self.currentPosition

        if (
            self.currentPosition == self.upperBound
            or self.currentPosition == self.lowerBound
        ):
            return "{0}/{1}/{2}".format(self.name, "position", "limit")
        else:
            return "{0}/{1}/{2}".format(self.name, "position", self.currentPosition)

    def move_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "move")
        return int(params[0])

    def goto(self, position):
        print(position)
        endPoint = self.refPoints[position]
        self.move(endPoint - self.currentPosition)

    def goto_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "goto")
        return params[0]

    def degMove_parser(self, params):
        try:
            steps = float(params[0])
        except ValueError:
            raise ArgumentError(self.name, "degMove", params)
        return steps

    def degMove(self, deg):
        print("{0} degrees".format(deg))
        step = deg / self.degPerStep
        step = int(step * self.gearRatio)
        step = round(step)
        response = self.move(step)
        return response

    def homeMove(self, stepLimit=5000, additionalSteps=10):
        if self.currentPosition > 0:
            direction = -1
        elif self.currentPosition < 0:
            direction = 1
        else:
            return

        self.move(direction * stepLimit)
        self.move(direction * additionalSteps)

    def reset(self):
        if self.homeSwitch is not None:
            self.home(1)
        else:
            self.move(-self.currentPosition)
        # pass

    def home(self, params):
        if self.homeSwitch is not None:
            self.homing = True
            self.customHome(self)
            self.homing = False
        else:
            print("No homing switch is attached to this motor")

    def customHome(self, motor):
        pass


class PololuDCMotor(BaseController):
    deviceType = "controller"

    def __init__(
        self,
        name,
        pwmPin,
        directionPin,
        notEnablePin,
        stopPin=None,
        rising=True,
        steadyState=20000,
        frequency=100,
        dutyCycle=0,
        pwmScaler=255,
    ):
        super().__init__(name)
        self.pwmPin = pwmPin
        self.directionPin = directionPin
        self.notEnablePin = notEnablePin
        self.frequency = frequency
        self.dutyCycle = dutyCycle
        self.stopPin = stopPin
        self.steadyState = steadyState
        self.pulseCount = 0
        self.pwmScaler = pwmScaler

        if self.stopPin is not None:
            if rising:
                pi.set_mode(stopPin, pigpio.INPUT)
                pi.set_pull_up_down(stopPin, pigpio.PUD_DOWN)
                pi.set_glitch_filter(stopPin, self.steadyState)
                pi.callback(stopPin, pigpio.RISING_EDGE, self.__stop)
                # gpio.setup(self.stopPin, gpio.IN, pull_up_down=gpio.PUD_DOWN)
                # gpio.add_event_detect(self.stopPin, gpio.RISING, callback=self.__stop, bouncetime=100)
            else:
                pi.set_mode(stopPin, pigpio.INPUT)
                pi.set_pull_up_down(stopPin, pigpio.PUD_UP)
                pi.set_glitch_filter(stopPin, self.steadyState)
                pi.callback(stopPin, pigpio.FALLING_EDGE, self.__stop)
                # gpio.setup(self.stopPin, gpio.IN, pull_up_down=gpio.PUD_UP)
                # gpio.add_event_detect(self.stopPin, gpio.FALLING, callback=self.__stop, bouncetime=100)

        # gpio.setup([self.pwmPin, self.directionPin, self.notEnablePin], gpio.OUT)
        for pin in [self.pwmPin, self.directionPin, self.notEnablePin]:
            pi.set_mode(pin, pigpio.OUTPUT)
        # gpio.output(self.notEnablePin, gpio.LOW)
        pi.write(self.notEnablePin, 0)
        pi.set_PWM_frequency(self.pwmPin, self.frequency)
        pi.set_PWM_dutycycle(self.pwmPin, dutyCycle)
        # self.pwm = gpio.PWM(self.pwmPin, self.frequency)
        # self.pwm.start(dutyCycle)

        self.state = {}

    def __stop(self, gpio, level, tick):
        print("MOTOR IS CRASHING! HALTING!")
        self.throttle(self.throttle_parser([1]))
        time.sleep(2)
        self.throttle(self.throttle_parser([0]))
        pi.stop()
        sys.exit(0)
        # self.pulseCount += 1
        # print("PULSED {0}".format(self.pulseCount))
        # gpio.cleanup()
        # sys.exit(0)

    def throttle(self, speed):
        # if speed >= 0:
        #     # gpio.output(self.directionPin, gpio.LOW)
        # else:
        #     # gpio.output(self.directionPin, gpio.HIGH)
        pi.write(self.directionPin, speed <= 0)

        self.dutyCycle = abs(speed)
        # self.pwm.ChangeDutyCycle(self.dutyCycle)
        pi.set_PWM_dutycycle(self.pwmPin, self.dutyCycle)

    def throttle_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "throttle")

        speed = float(params[0]) * self.pwmScaler

        if speed < -self.pwmScaler or speed > self.pwmScaler:
            raise ArgumentError(
                self.name, "throttle", speed / self.pwmScaler, "-1 <= speed <= 1"
            )

        return speed

    def reset(self):
        pass


class ArduCamMultiCamera(BaseController):
    deviceType = "measurement"

    def __init__(
        self,
        name,
        numCameras,
        videoNumber=0,
        defaultSettings=None,
        i2cbus=11,
        initialCamera="a",
        controlPins=[4, 17, 18],
        cameraNamesDict=None,
    ):
        super().__init__(name)
        self.videoNumber = videoNumber
        self.numCameras = numCameras
        self.experiment = None
        self.state = {}
        self.defaultSettings = defaultSettings
        self.i2cbus = i2cbus
        self.cameraNames = cameraNamesDict
        self.initialCamera = initialCamera

        # Define Pins
        # Board Pin 7 = BCM Pin 4 = Selection
        # Board Pin 11 = BCM Pin 17 = Enable 1
        # Board Pin 12 = BCM Pin 18 = Enable 2
        # See Arducam User Guide https://www.uctronics.com/download/Amazon/B0120.pdf
        self.selection, self.enable1, self.enable2 = controlPins
        self.channels = controlPins
        gpio.setup(self.channels, gpio.OUT)

        self.cameraDict = {
            "a": (gpio.LOW, gpio.LOW, gpio.HIGH),
            "b": (gpio.HIGH, gpio.LOW, gpio.HIGH),
            "c": (gpio.LOW, gpio.HIGH, gpio.LOW),
            "d": (gpio.HIGH, gpio.HIGH, gpio.LOW),
            "off": (gpio.LOW, gpio.HIGH, gpio.HIGH),
        }

        self.camerai2c = {
            "a": "i2cset -y {0} 0x70 0x00 0x04".format(self.i2cbus),
            "c": "i2cset -y {0} 0x70 0x00 0x06".format(self.i2cbus),
            "d": "i2cset -y {0} 0x70 0x00 0x07".format(self.i2cbus),
            "b": "i2cset -y {0} 0x70 0x00 0x05".format(self.i2cbus),
        }

        # Set camera for A
        _cameraNames = ["a", "b", "c", "d"]
        for i in range(self.numCameras):
            self.camera(_cameraNames[i])
            time.sleep(5)
        if initialCamera in _cameraNames:
            self.camera(initialCamera)
        else:
            self.cameraName(initialCamera)
        time.sleep(5)

    def camera(self, param):
        # Param should be a, b, c, d, or off
        print("Switching to camera " + param)
        os.system(self.camerai2c[param])
        gpio.output(self.channels, self.cameraDict[param])

    def camera_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "camera")
        param = params[0].lower()
        if param not in self.cameraDict:
            raise ArgumentError(self.name, "camera", param, ["a", "b", "c", "d", "off"])
        return params[0].lower()

    # This is a translation layer so we can switch by named camera instead of slots.
    # The name should be translated from the dictionary self.cameraNames
    def cameraName(self, param):
        cameraSlot = self.cameraNames[param]
        print("Switching to camera {0}, slot {1}".format(param, cameraSlot))
        os.system(self.camerai2c[cameraSlot])
        gpio.output(self.channels, self.cameraDict[cameraSlot])

    def cameraName_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "cameraName")
        param = params[0].lower()
        if param not in self.cameraNames:
            raise ArgumentError(self.name, "cameraName", param, self.cameraNames)
        return param

    def imageMod(self, params):
        imageControl = params[0]
        controlValue = params[1]
        subprocess.run(
            "v4l2-ctl -d /dev/video{0} -c {1}={2}".format(
                self.videoNumber, imageControl, controlValue
            ),
            shell=True,
        )

    def imageMod_parser(self, params):
        if len(params) != 2:
            raise ArgumentNumberError(len(params), 2, "imageMod")
        return params

    def reset(self):
        self.camera(self.initialCamera)
        if self.defaultSettings is not None:
            for setting, value in self.defaultSettings.items():
                self.imageMod([setting, value])
                time.sleep(0.1)


class ElectronicScreen(BaseController):
    deviceType = "controller"

    def __init__(self, name, pin):
        super().__init__(name)
        self.pin = pin
        self.state = "off"
        gpio.setup(self.pin, gpio.OUT)

    def on(self, params):
        gpio.output(self.pin, gpio.HIGH)

    def off(self, params):
        gpio.output(self.pin, gpio.LOW)

    def reset(self):
        gpio.output(self.pin, gpio.LOW)


class LimitSwitch(BaseController):
    deviceType = "controller"

    def __init__(self, name, pin, state=False):
        super().__init__(name)
        self.pin = pin
        self.state = state
        gpio.setup(self.pin, gpio.IN, pull_up_down=gpio.PUD_DOWN)

    def getStatus(self, params):
        state = gpio.input(self.pin)
        self.state = state
        return state

    def switchAction(self, motor, steps):
        pass

    def reset(self):
        pass


class HomeSwitch(BaseController):
    deviceType = "controller"

    def __init__(self, name, pin, state=False):
        super().__init__(name)
        self.pin = pin
        self.state = state
        gpio.setup(self.pin, gpio.IN, pull_up_down=gpio.PUD_DOWN)

    def getStatus(self, params):
        state = gpio.input(self.pin)
        self.state = state
        return state

    def reset(self):
        pass


class SingleGPIO(BaseController):
    deviceType = "controller"

    def __init__(self, name, pin, initialState=False):
        super().__init__(name)
        self.pin = pin
        gpio.setup(self.pin, gpio.OUT)
        if initialState:
            self.state = "off"
            gpio.output(self.pin, gpio.HIGH)
        else:
            self.state = "off"
            gpio.output(self.pin, gpio.LOW)

    def on(self, params):
        gpio.output(self.pin, gpio.HIGH)

    def off(self, params):
        gpio.output(self.pin, gpio.LOW)

    def reset(self):
        gpio.output(self.pin, gpio.LOW)


class PushButton(BaseController):
    deviceType = "controller"

    def __init__(self, name, pin, initialState=False, delay=0.1):
        super().__init__(name)
        self.pin = pin
        self.delay = delay
        self.initialState = initialState
        gpio.setup(self.pin, gpio.OUT)
        if initialState:
            self.state = "on"
            gpio.output(self.pin, gpio.HIGH)
        else:
            self.state = "off"
            gpio.output(self.pin, gpio.LOW)

    def press(self, params):
        if self.initialState:
            gpio.output(self.pin, gpio.LOW)
            time.sleep(self.delay)
            gpio.output(self.pin, gpio.HIGH)
        else:
            gpio.output(self.pin, gpio.HIGH)
            time.sleep(self.delay)
            gpio.output(self.pin, gpio.LOW)

    def reset(self):
        gpio.output(self.pin, gpio.LOW)


class PWMChannel(BaseController):
    deviceType = "controller"

    def __init__(self, name, pin, frequency, defaultDutyCycle=0):
        super().__init__(name)
        self.pin = pin
        self.frequency = frequency
        self.defaultDutyCycle = defaultDutyCycle
        self.dutyCycle = defaultDutyCycle
        gpio.setup(self.pin, gpio.OUT)
        self.pwm = gpio.PWM(self.pin, self.frequency)
        self.pwm.start(self.dutyCycle)
        self.state = {}

    def power(self, dutyCycle):
        self.pwm.ChangeDutyCycle(dutyCycle)

    def power_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "power")
        dutyCycle = float(params[0])
        if dutyCycle < 0 or dutyCycle > 100:
            raise ArgumentError(
                self.name, "power", dutyCycle, allowed="0 <= dutyCycle <= 100"
            )
        return dutyCycle

    def reset(self):
        self.pwm.ChangeDutyCycle(self.defaultDutyCycle)


class S42CStepperMotor(BaseController):
    """
    !There are no argument type check for this class!
    TODO Barry (chutian_wang@ucsb.edu) if checks became necessary.

    args:
        name -- str: the name of this motor.
        EN -- int: the enable pin port.
        STEP -- int: the step pin port.
        DIR -- int: the direction pin port.
        bounds -- tuple: the limiting step bounds beyond which the
            motor should not reach, (LB, UB). Set to (None, None)
            If there are no restrictions (NOT RECOMMENDED). LB
            and UB are steps relative to initial position. If
            LB > 0 or UB < 0 initializaition will fail.

    kwargs:
        refPoints -- dict({}): A table of predefined reference points
            that can be quickly accessed by calling the `goto()`
            function. {str: int}
        delay -- float(0.02): the step pulse period in seconds.
        microstep -- int(2): the microstep setting on the motor.
            Choose a higher value if more precision is needed.
            This parameter is currently NOT USED. TODO Barry
            if it needs to be implemented.
        pi -- any(None): the host Rpi pigpio.pi instance. Default None
            will use the local host pi.

    The S24CStepperMotor shares the same controls as the StepperI2C motor.
    It has the following `microStep` settings on the OLED menu:

        2 -- 400  steps per rev or 0.9 degrees per microstep.
        4 -- 800  steps per rev or 0.45 degrees per microstep.
        8 -- 1600 steps per rec or 0.225 degrees per microstep.

    Hihger microsteps are 16, 32, 64, 128, 256.
    Note: more steps usually results in less torque but higher precision.

    Bound crossing handling:
        If any bounds are hit, the original move steps will be registered
        internally but the physical motor will not pass through the bound.
        If curPos is not within the bounds, any moves that attempt to push
        the motor further from the bounds range will not have physical
        effect.
        e.g. UB = 100, curPos = 0, move(101) will set curPos to 101 and
        send 100 step pulses to the motor.
        e.g. UB = 100, curPos = 200, move(10) will set curPos to 210 and
        send 0 step pulses to the motor.
        e.g. UB = 100, curPos = 200, move(-150) will set curPos to 50 and
        send 50 step pulses to the motor.

    Multithreading/Callback support:
        None, TODO barry if required in the future.
    """

    deviceType = "controller"

    def __init__(
        self,
        name: str,
        EN: int,
        STEP: int,
        DIR: int,
        bounds: tuple,
        refPoints={},
        microstep=2,
        gearRatio=1,
        _pi=None,
        stepWaitTime=0,
    ):
        super().__init__(name)
        self.EN = EN
        self.STEP = STEP
        self.DIR = DIR
        self.LB = bounds[0]
        self.UB = bounds[1]
        # Bounds warning logic, remove if unecessary to improve performance
        if self.LB == None or self.UB == None:
            if self.LB == None and self.UB == None:
                warn(self.name + " is initialized without any bounds!", RuntimeWarning)
            elif self.LB == None:
                warn(
                    self.name + " is initialized without a lower bound.", RuntimeWarning
                )
            elif self.UB == None:
                warn(
                    self.name + " is initialized without a upper bound.", RuntimeWarning
                )
        else:
            if self.LB > self.UB:
                warn(self.name + " is initialized with LB > UB!", RuntimeWarning)
            if self.LB > 0:
                warn(
                    self.name + " is initialized with a positive lower bound.",
                    RuntimeWarning,
                )
            if self.UB < 0:
                warn(
                    self.name + " is initialized with a negative lower bound.",
                    RuntimeWarning,
                )
        self.refPoints = refPoints
        self.delay = 0.0001
        self.mStep = microstep
        self.gearRatio = gearRatio
        self.stepWaitTime = stepWaitTime
        # The host Rpi, in case multihost systems are used in the future
        if not _pi:
            self.pi = pi
        else:
            self.pi = _pi
            warn("Foreign controls may not be supported.", RuntimeWarning)

        # The home position (TODO future update)
        self.homePos = 0
        # The ideal current position.
        self.curPos = 0

        self.pi.set_mode(self.STEP, pigpio.OUTPUT)
        self.pi.set_pull_up_down(self.STEP, pigpio.PUD_DOWN)
        self.pi.set_mode(self.EN, pigpio.OUTPUT)
        self.pi.set_pull_up_down(self.EN, pigpio.PUD_UP)
        self.pi.set_mode(self.DIR, pigpio.OUTPUT)
        self.pi.set_pull_up_down(self.DIR, pigpio.PUD_DOWN)

        # Always enable closed loop control. EN is active low.
        self.pi.write(self.EN, 0)
        self.state = {"position": self.curPos}

    def move_parser(self, params: list) -> int:
        """
        args:
            params -- iterable(list): The parameter list that contains cmd strings.

        ret:
            int: The number of steps to move.

        Will throw ArgumentNumberError if len(params) != 1.
        """
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "move")
        return int(params[0])

    def goto_parser(self, params: list):
        """
        args:
            params -- iterable(list): The parameter list that contains cmd strings.

        ret:
            str: The target refPoint.

        Will throw ArgumentNumberError if len(params) != 1.
        """
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "goto")
        return params[0]

    def degMove_parser(self, params: list):
        """
        args:
            params -- iterable(list): The parameter list that contains cmd strings.

        ret:
            str: The target refPoint.

        Will throw ArgumentNumberError if len(params) != 1.
        """
        try:
            deg = float(params[0])
        except ValueError:
            raise ArgumentError(self.name, "degMove", params)
        return deg

    def __sqGenPWM(self, pin, period, num):
        pOn = pigpio.pulse(1 << self.STEP, 0, int(self.delay * 1e6 / 2))
        pOff = pigpio.pulse(0, 1 << self.STEP, int(self.delay * 1e6 / 2))
        pulse = [pOn, pOff]
        pi.wave_clear()
        pi.wave_add_generic(pulse)
        stepWave = pi.wave_create()

        absteps = abs(num)
        step_y = absteps // 256
        step_x = absteps % 256

        pi.wave_chain([255, 0, stepWave, 255, 1, step_x, step_y])

        while pi.wave_tx_busy():
            time.sleep(0.1)

    def move(self, steps: int):
        """
        args:
            steps -- int: The desired number of steps to move.

        ret:
            str: A response string ("{self.name}/position/{self.state["position"]}"
            or "{self.name}/position/limit").

        Attemps to move `steps` number of steps. Will stop if any bounds are hit. See
        class documentation `Bound crossing handling` for details.
        """
        # bounds checking decision tree.
        # optimize this with LUT if better performance is desiered
        targetPos = self.curPos + steps

        if self.LB == None and self.UB == None:
            moveSteps = steps

        elif self.LB == None:
            if self.curPos < self.UB:
                if targetPos <= self.UB:
                    moveSteps = targetPos - self.curPos
                else:  # targetPos > self.UB
                    moveSteps = self.UB - self.curPos

            if self.curPos >= self.UB:
                if targetPos <= self.UB:
                    moveSteps = targetPos - self.UB
                else:  # targetPos > self.UB
                    moveSteps = 0

        elif self.UB == None:
            if self.curPos <= self.LB:
                if targetPos < self.LB:
                    moveSteps = 0
                else:  # targetPos >= self.LB
                    moveSteps = targetPos - self.LB

            if self.curPos > self.LB:
                if targetPos < self.LB:
                    moveSteps = self.LB - self.curPos
                else:  # targetPos >= self.LB:
                    moveSteps = targetPos - self.curPos

        else:
            if self.curPos <= self.LB:
                if targetPos < self.LB:
                    moveSteps = 0
                elif targetPos >= self.LB and targetPos <= self.UB:
                    moveSteps = targetPos - self.LB
                else:  # targetPos > self.UB
                    moveSteps = self.UB - self.LB

            if self.curPos > self.LB and self.curPos < self.UB:
                if targetPos < self.LB:
                    moveSteps = self.LB - self.curPos
                elif targetPos >= self.LB and targetPos <= self.UB:
                    moveSteps = targetPos - self.curPos
                else:  # targetPos > self.UB
                    moveSteps = self.UB - self.curPos

            if self.curPos >= self.UB:
                if targetPos < self.LB:
                    moveSteps = self.LB - self.UB
                elif targetPos >= self.LB and targetPos <= self.UB:
                    moveSteps = targetPos - self.UB
                else:  # targetPos > self.UB
                    moveSteps = 0

        # set DIR pin
        if moveSteps >= 0:
            self.pi.write(self.DIR, 1)
        else:
            self.pi.write(self.DIR, 0)

        self.__sqGenPWM(self.STEP, self.delay, abs(moveSteps))
        time.sleep(self.stepWaitTime * abs(moveSteps))

        # update self.curPos and self.state
        self.curPos = targetPos
        self.state["position"] += moveSteps

        if self.state["position"] == self.UB or self.state["position"] == self.LB:
            return "{0}/{1}/{2}".format(self.name, "position", "limit")
        else:
            return "{0}/{1}/{2}".format(self.name, "position", self.state["position"])

    def goto(self, ref: str, safe=True):
        """
        args:
            ref -- int: The target refPoint.
            safe -- bool(True): If `safe` is set True, the motor will not attempt
                to move to a refpoint beyond the bounds. Otherwise it will stop at
                the bound.

        ret:
            str: A response string ("{self.name}/position/{self.state["position"]}"
            or "{self.name}/position/limit").

        Attemps to move to the target refPoint. Will stop if any bounds are hit. See
        class documentation `Bound crossing handling` for details.
        """
        if safe and (self.refPoints[ref] > self.UB or self.refPoints[ref] < self.LB):
            warn(self.name + " refPoint out of bounds, move canceled.", RuntimeWarning)
            return "{0}/{1}/{2}".format(self.name, "position", self.state["position"])
        else:
            target = self.refPoints[ref]
            return self.move(target - self.curPos)

    def reset(self):
        return self.move(-self.curPos)


class FS5103RContinuousMotor(BaseController):
    deviceType = "controller"

    def __init__(self, name, PWM, limitPin=None, _reversed=False, pi=None):
        """
        This is the controller for the FS5103R continuous motor used on the gamma labs.
        When active (default, not disabled) this motor will use one of the PWM channels.
        This controller is made compatable with any code that uses the same commands as
        the PololuDCMotor. It also has an additional `disable` method that will stop the
        PWM output and release the PWM channel. The motor will stop as soon as a limit
        is triggered (limit pin pulled high).
        args:
            name -- str: the name of this motor.
            PWM -- int: the PWM pin for the motor.

        kwargs:
            pi -- any(None): the host Rpi pigpio.pi instance. Default None
                will use the local host pi.
            limitPin -- int/(None): the limit interrupt pin for the motor. Default None will
                not have any limit protection.
            _reversed -- bool(False): if the motor's direction should be _reversed set this to True.
        """
        super().__init__(name)
        self.PWM = PWM
        self.limitPin = limitPin
        self._reversed = _reversed
        # The host Rpi, in case multihost systems are used in the future
        if not pi:
            self.pi = pigpio.pi()
        else:
            self.pi = pi
            warn("Foreign controls may not be supported.", RuntimeWarning)

        self.pi.set_mode(self.PWM, pigpio.OUTPUT)
        self.pi.set_pull_up_down(self.PWM, pigpio.PUD_DOWN)

        # Set the callback trigger pin, the callback function, and the
        # glitch filter for the limit switch
        if self.limitPin != None:
            self.pi.set_mode(self.limitPin, pigpio.INPUT)
            self.pi.set_glitch_filter(limitPin, 50)
            self.stopCallback = self.pi.callback(
                limitPin, pigpio.RISING_EDGE, self.__stop
            )
            self.startCallback = self.pi.callback(
                limitPin, pigpio.FALLING_EDGE, self.__resetCallback
            )
            self.pi.set_mode(self.PWM, pigpio.INPUT)
            self.pi.set_pull_up_down(self.PWM, pigpio.PUD_DOWN)
        self.state = {"running": False, "throttle": 0}

    def throttle(self, throttle):
        """
        args:
            throttle -- int: desired percent speed in the range [-1.0, 1.0],
                1.0 sets the speed fully forwards and -1.0 sets the speed
                fully backwards. 0 is stall.
        """
        self.state["running"] = True
        self.pi.set_mode(self.PWM, pigpio.OUTPUT)
        dutyCycle = int(1500 + throttle * 1000)
        self.pi.set_servo_pulsewidth(self.PWM, dutyCycle)
        self.state["throttle"] = throttle
        return self.state

    def __stop(self, gpio, level, tick):
        self.stopCallback.cancel()
        self.startCallback = self.pi.callback(
            self.limitPin, pigpio.FALLING_EDGE, self.__resetCallback
        )
        # print(self.name, f"has reached its limit. status {level}")
        self.throttle(0)

    def __resetCallback(self, gpio, level, tick):
        self.startCallback.cancel()
        self.stopCallback = self.pi.callback(
            self.limitPin, pigpio.RISING_EDGE, self.__stop
        )
        # print(self.name, f"stop callback reset. status {level}")

    def disable(self):
        self.pi.set_mode(self.PWM, pigpio.INPUT)
        self.state["running"] = False
        self.state["throttle"] = 0.0
        return self.state

    def throttle_parser(self, params):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "setSpeed")
        throttle = float(params[0])
        if self._reversed:
            throttle = -throttle
        if throttle < -1 or throttle > 1:
            raise ArgumentError(
                self.name, "throttle", throttle, allowed="-1 <= throttle <= 1"
            )
        return throttle

    def reset(self):
        pass


class GeneralPWMServo(BaseController):
    deviceType = "controller"

    def __init__(self, name, PWM, pi=None):
        """
        This is a control class for generic PWM controlled 180 degrees
        servo.
        args:
            name -- str: the name of this servo.
            PWM -- int: the PWM pin for the servo.

        kwargs:
            pi -- any(None): the host Rpi pigpio.pi instance. Default None
                will use the local host pi.
        """
        super().__init__(name)
        self.PWM = PWM
        if not pi:
            self.pi = pigpio.pi()
        else:
            self.pi = pi
            warn("Foreign controls may not be supported.", RuntimeWarning)

        self.pi.set_mode(self.PWM, pigpio.OUTPUT)
        self.pi.set_pull_up_down(self.PWM, pigpio.PUD_DOWN)
        self.state = {"running": False, "angle": 0}

    def goto_parser(self, params: list):
        if len(params) != 1:
            raise ArgumentNumberError(len(params), 1, "goto")
        try:
            return float(params[0])
        except:
            raise ArgumentError(self.name, "goto", params[0], float)

    def goto(self, angle):
        """
        This method will use one of the PWM channels on the pi to control
        the servo.
        args:
            angle -- float: The target angle, this angle must be between +90
                and -90.
        """
        self.state["running"] = True
        self.pi.set_mode(self.PWM, pigpio.OUTPUT)
        if angle < -90 or angle > 90:
            raise ArgumentError(self.name, "goto", angle, allowed="-90 < angle < 90")
        dutyCycle = int(1500 + angle / 90 * 1000)
        self.pi.set_servo_pulsewidth(self.PWM, dutyCycle)
        self.state["angle"] = angle
        return self.state

    def disable(self):
        """
        This method will stop the servo control and free up the PWM channel.
        The servo will usually stop in place but this is technically undefined
        behavior.
        """
        self.pi.set_mode(self.PWM, pigpio.INPUT)
        self.state["running"] = False
        return self.state

    def reset(self):
        pass


class CommandError(Exception):
    def __init__(self, command, *args):
        self.command = command
        if args:
            self.message = args[0]
        else:
            self.message = "No command named '{0}' found".format(self.command)

    def __str__(self):
        return "CommandError, {0}".format(self.message)


class ArgumentNumberError(Exception):
    def __init__(self, total_args, allowed, command=None):
        self.total_args = total_args
        self.allowed = allowed
        self.command = command

    def __str__(self):
        if self.command is None:
            return "ArgumentNumberError, received {0} when {1} was expected.".format(
                self.total_args, self.allowed
            )
        else:
            return "ArgumentNumberError, command '{0}' received {1} when {2} was expected.".format(
                self.command, self.total_args, self.allowed
            )


class ArgumentError(Exception):
    def __init__(self, device_name, command, received, allowed=None):
        self.device_name = device_name
        self.command = command
        self.allowed = allowed
        self.received = received

    def __str__(self):
        if self.allowed is None:
            return "ArgumentError, Device, {0}, can't process command argument {1} by command {2}.".format(
                self.device_name, self.received, self.command
            )
        else:
            return "ArgumentError, Argument {0}, is not one of the allowed commands, {1}, for device, {2}, running command {3}.".format(
                self.received, self.allowed, self.device_name, self.command
            )
